#!/usr/bin/python3 -BbbEIsSttW all
# -*- coding: utf-8 -*-

"""
This is the main program of the "aminer" logfile miner tool.
It does not import any local default site packages to decrease the attack surface due to manipulation of unused but available packages.

CAVEAT: This process will keep running with current permissions, no matter what was specified in 'AminerUser' and 'AminerGroup'
configuration properties. This is required to allow the aminer parent parent process to reopen log files, which might need the
elevated privileges.

NOTE: This tool is developed to allow secure operation even in hostile environment, e.g. when one directory, where aminer attempts
to open logfiles is already under full control of an attacker. However it is not intended to be run as SUID-binary, this would
require code changes to protect also against standard SUID attacks.

Parameters:
* --config [file]: Location of configuration file, defaults to '/etc/aminer/config.py' when not set.
* --run-analysis: This parameters is NOT intended to be used on command line when starting aminer, it will trigger execution
  of the unprivileged aminer background child performing the real analysis.

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have received a copy
of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import errno
import os
import re
import socket
import time
import sys
import logging
import shutil
import warnings
import argparse
import stat
import tempfile
import ast
from pwd import getpwnam
from grp import getgrnam
from logging.handlers import RotatingFileHandler

# As site packages are not included, define from where we need to execute code before loading it.
sys.path = sys.path[1:] + ['/usr/lib/logdata-anomaly-miner', '/etc/aminer/conf-enabled']
import aminer.AminerConfig as AminerConfig  # skipcq: FLK-E402
from aminer.util.StringUtil import colflame, flame, supports_color, decode_string_as_byte_string  # skipcq: FLK-E402
from aminer.util.PersistenceUtil import clear_persistence, copytree  # skipcq: FLK-E402
from aminer.util import SecureOSFunctions  # skipcq: FLK-E402
from aminer.AnalysisChild import AnalysisChild  # skipcq: FLK-E402
from aminer.input.LogStream import FileLogDataResource, UnixSocketLogDataResource  # skipcq: FLK-E402
from metadata import __version_string__, __version__  # skipcq: FLK-E402


child_termination_triggered_flag = False
offline_mode = False


def run_analysis_child(aminer_config, program_name):
    """Run the Analysis Child."""
    # Verify existence and ownership of persistence directory.
    logging.getLogger(AminerConfig.REMOTE_CONTROL_LOG_NAME).info('aminer started.')
    logging.getLogger(AminerConfig.DEBUG_LOG_NAME).info('aminer started.')
    persistence_dir_name = aminer_config.config_properties.get(AminerConfig.KEY_PERSISTENCE_DIR, AminerConfig.DEFAULT_PERSISTENCE_DIR)
    persistence_dir_fd = SecureOSFunctions.secure_open_base_directory(persistence_dir_name, os.O_RDONLY | os.O_DIRECTORY | os.O_PATH)
    stat_result = os.fstat(persistence_dir_fd)
    if ((not stat.S_ISDIR(stat_result.st_mode)) or ((stat_result.st_mode & stat.S_IRWXU) != 0o700) or (
            stat_result.st_uid != os.getuid()) or (stat_result.st_gid != os.getgid())):
        msg = 'FATAL: persistence directory "%s" has to be owned by analysis process (uid %d!=%d, gid %d!=%d) and have access mode 0700 ' \
              'only!' % (repr(persistence_dir_name), stat_result.st_uid, os.getuid(), stat_result.st_gid, os.getgid())
        print(msg, file=sys.stderr)
        logging.getLogger(AminerConfig.DEBUG_LOG_NAME).critical(msg)
        sys.exit(1)
    import posix1e
    # O_PATH is problematic when checking ACL. However it is possible to check the ACL using the file name.
    try:
        if posix1e.has_extended(persistence_dir_name):
            msg = 'WARNING: SECURITY: Extended POSIX ACLs are set in %s, but not supported by the aminer. Backdoor access could be ' \
                  'possible.' % persistence_dir_name.decode()
            print(msg, file=sys.stderr)
            logging.getLogger(AminerConfig.DEBUG_LOG_NAME).warning(msg)
    except OSError:  # system does not support POSIX ACLs.
        pass

    child = AnalysisChild(program_name, aminer_config)
    child.offline_mode = offline_mode
    # This function call will only return on error or signal induced normal termination.
    child_return_status = child.run_analysis(3)
    if child_return_status == 0:
        sys.exit(0)
    msg = '%s: run_analysis terminated with unexpected status %d' % (program_name, child_return_status)
    print(msg, file=sys.stderr)
    logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)
    sys.exit(1)


def initialize_loggers(aminer_config, aminer_user_id, aminer_grp_id):
    """Initialize all loggers."""
    datefmt = '%d/%b/%Y:%H:%M:%S %z'

    log_dir = aminer_config.config_properties.get(AminerConfig.KEY_LOG_DIR, AminerConfig.DEFAULT_LOG_DIR)
    if log_dir == AminerConfig.DEFAULT_LOG_DIR:
        try:
            if not os.path.isdir(log_dir):
                persistence_dir_path = aminer_config.config_properties.get(
                    AminerConfig.KEY_PERSISTENCE_DIR, AminerConfig.DEFAULT_PERSISTENCE_DIR)
                persistence_dir_fd = SecureOSFunctions.secure_open_base_directory(
                    persistence_dir_path, os.O_RDONLY | os.O_DIRECTORY | os.O_PATH)
                if SecureOSFunctions.base_dir_path.decode() == AminerConfig.DEFAULT_PERSISTENCE_DIR:
                    relative_path_log_dir = os.path.split(AminerConfig.DEFAULT_LOG_DIR)[1]
                    os.mkdir(relative_path_log_dir, dir_fd=persistence_dir_fd)
                    os.chown(relative_path_log_dir, aminer_user_id, aminer_grp_id, dir_fd=persistence_dir_fd, follow_symlinks=False)
        except OSError as e:
            if e.errno != errno.EEXIST:
                msg = 'Unable to create log-directory: %s' % log_dir
            else:
                msg = e
            logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg.strip('\n'))
            print(msg, file=sys.stderr)

    tmp_value = aminer_config.config_properties.get(AminerConfig.KEY_REMOTE_CONTROL_LOG_FILE)
    if tmp_value is not None and b'/' in tmp_value:
        print('%s attribute must not contain a full directory path, but only the filename.' % AminerConfig.KEY_REMOTE_CONTROL_LOG_FILE,
              file=sys.stderr)
        sys.exit(1)
    tmp_value = aminer_config.config_properties.get(AminerConfig.KEY_STAT_LOG_FILE)
    if tmp_value is not None and b'/' in tmp_value:
        print('%s attribute must not contain a full directory path, but only the filename.' % AminerConfig.KEY_STAT_LOG_FILE,
              file=sys.stderr)
        sys.exit(1)
    tmp_value = aminer_config.config_properties.get(AminerConfig.KEY_DEBUG_LOG_FILE)
    if tmp_value is not None and b'/' in tmp_value:
        print('%s attribute must not contain a full directory path, but only the filename.' % AminerConfig.KEY_DEBUG_LOG_FILE,
              file=sys.stderr)
        sys.exit(1)

    max_bytes = aminer_config.config_properties.get(AminerConfig.KEY_LOG_ROTATION_MAX_BYTES, AminerConfig.DEFAULT_LOG_ROTATION_MAX_BYTES)
    backup_count = aminer_config.config_properties.get(
        AminerConfig.KEY_LOG_ROTATION_BACKUP_COUNT, AminerConfig.DEFAULT_LOG_ROTATION_BACKUP_COUNT)
    log_dir_fd = SecureOSFunctions.secure_open_log_directory(log_dir, os.O_RDONLY | os.O_DIRECTORY | os.O_PATH)
    rc_logger = logging.getLogger(AminerConfig.REMOTE_CONTROL_LOG_NAME)
    rc_logger.setLevel(logging.DEBUG)
    remote_control_log_file = aminer_config.config_properties.get(
        AminerConfig.KEY_REMOTE_CONTROL_LOG_FILE, os.path.join(log_dir, AminerConfig.DEFAULT_REMOTE_CONTROL_LOG_FILE))
    if not remote_control_log_file.startswith(log_dir):
        remote_control_log_file = os.path.join(log_dir, remote_control_log_file)
    try:
        rc_file_handler = RotatingFileHandler(remote_control_log_file, maxBytes=max_bytes, backupCount=backup_count)
        os.chown(remote_control_log_file, aminer_user_id, aminer_grp_id, dir_fd=log_dir_fd, follow_symlinks=False)
    except OSError as e:
        print('Could not create or open %s: %s. Stopping..' % (remote_control_log_file, e), file=sys.stderr)
        sys.exit(1)
    rc_file_handler.setFormatter(logging.Formatter(fmt='%(asctime)s %(levelname)s %(message)s', datefmt=datefmt))
    rc_logger.addHandler(rc_file_handler)
    logging.addLevelName(15, "REMOTECONTROL")

    stat_logger = logging.getLogger(AminerConfig.STAT_LOG_NAME)
    stat_logger.setLevel(logging.INFO)
    stat_log_file = aminer_config.config_properties.get(
        AminerConfig.KEY_STAT_LOG_FILE, os.path.join(log_dir, AminerConfig.DEFAULT_STAT_LOG_FILE))
    if not stat_log_file.startswith(log_dir):
        stat_log_file = os.path.join(log_dir, stat_log_file)
    try:
        stat_file_handler = RotatingFileHandler(stat_log_file, maxBytes=max_bytes, backupCount=backup_count)
        os.chown(stat_log_file, aminer_user_id, aminer_grp_id, dir_fd=log_dir_fd, follow_symlinks=False)
    except OSError as e:
        print('Could not create or open %s: %s. Stopping..' % (stat_log_file, e), file=sys.stderr)
        sys.exit(1)
    stat_file_handler.setFormatter(logging.Formatter(fmt='%(asctime)s %(message)s', datefmt=datefmt))
    stat_logger.addHandler(stat_file_handler)

    debug_logger = logging.getLogger(AminerConfig.DEBUG_LOG_NAME)
    if AminerConfig.DEBUG_LEVEL == 0:
        debug_logger.setLevel(logging.ERROR)
    elif AminerConfig.DEBUG_LEVEL == 1:
        debug_logger.setLevel(logging.INFO)
    else:
        debug_logger.setLevel(logging.DEBUG)
    debug_log_file = aminer_config.config_properties.get(
        AminerConfig.KEY_DEBUG_LOG_FILE, os.path.join(log_dir, AminerConfig.DEFAULT_DEBUG_LOG_FILE))
    if not debug_log_file.startswith(log_dir):
        debug_log_file = os.path.join(log_dir, debug_log_file)
    try:
        debug_file_handler = RotatingFileHandler(debug_log_file, maxBytes=max_bytes, backupCount=backup_count)
        os.chown(debug_log_file, aminer_user_id, aminer_grp_id, dir_fd=log_dir_fd, follow_symlinks=False)
    except OSError as e:
        print('Could not create or open %s: %s. Stopping..' % (debug_log_file, e), file=sys.stderr)
        sys.exit(1)
    debug_file_handler.setFormatter(logging.Formatter(fmt='%(asctime)s %(levelname)s %(message)s', datefmt=datefmt))
    debug_logger.addHandler(debug_file_handler)


def parse_var(s):
    """
    Parse a key, value pair, separated by "=".
    That's the reverse of ShellArgs.

    On the command line (argparse) a declaration will typically look like:
        foo=hello
    or
        foo="hello world"
    """
    items = s.split("=")
    key = items[0].strip()  # we remove blanks around keys, as is logical
    if len(items) > 1:
        # rejoin the rest:
        value = "=".join(items[1:])
    return key, value


def parse_vars(items):
    """Parse a series of key-value pairs and return a dictionary."""
    d = {}

    if items:
        for item in items:
            key, value = parse_var(item)
            d[key] = value
    return d


def main():
    """Run the aminer main program."""
    # Extract program name, but only when sure to contain no problematic characters.
    warnings.filterwarnings('ignore', category=ImportWarning)
    program_name = sys.argv[0].split('/')[-1]
    if (program_name == '.') or (program_name == '..') or (re.match('^[a-zA-Z0-9._-]+$', program_name) is None):
        print('Invalid program name, check your execution args', file=sys.stderr)
        sys.exit(1)

    # We will not read stdin from here on, so get rid of it immediately, thus aberrant child cannot manipulate caller's stdin using it.
    stdin_fd = os.open('/dev/null', os.O_RDONLY)
    os.dup2(stdin_fd, 0)
    os.close(stdin_fd)

    help_message = 'aminer - logdata-anomaly-miner\n'
    if supports_color():
        help_message += colflame
    else:
        help_message += flame
    parser = argparse.ArgumentParser(description=help_message, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('-v', '--version', action='version', version=__version_string__)
    parser.add_argument('-u', '--check-updates', action='store_true', help='check if updates for the aminer are available.')
    parser.add_argument('-c', '--config', default='/etc/aminer/config.yml', type=str, help='path to the config-file')
    parser.add_argument('-D', '--daemon', action='store_false', help='run as a daemon process')
    parser.add_argument('-s', '--stat', choices=["0", "1", "2"], type=str,
                        help='set the stat level. Possible stat-levels are 0 for no statistics, 1 for normal statistic level and 2 for '
                             'verbose statistics.')
    parser.add_argument('-d', '--debug', choices=["0", "1", "2"], type=str,
                        help='set the debug level. Possible debug-levels are 0 for no debugging, 1 for normal output (INFO and above), 2 '
                             'for printing all debug information.')
    parser.add_argument('--run-analysis', action='store_true', help='enable/disable analysis')
    parser.add_argument('-C', '--clear', action='store_true', help='removes all persistence directories')
    parser.add_argument('-r', '--remove', action='append', type=str, help='removes a specific persistence directory')
    parser.add_argument('-R', '--restore', type=str, help='restore a persistence backup')
    parser.add_argument('-f', '--from-begin', action='store_true', help='removes RepositioningData before starting the aminer')
    parser.add_argument('-o', '--offline-mode', action='store_true', help='stop the aminer after all logs have been processed.')
    parser.add_argument("--config-properties", metavar="KEY=VALUE", nargs='+',
                        help="Set a number of config_properties by using key-value pairs (do not put spaces before or after the = sign). "
                             "If a value contains spaces, you should define it with double quotes: 'foo=\"this is a sentence\". Note that "
                             "values are always treated as strings. If values are already defined in the config_properties, the input "
                             "types are converted to the ones already existing.")

    args = parser.parse_args()

    if args.check_updates:
        import urllib3
        url = 'https://raw.githubusercontent.com/ait-aecid/logdata-anomaly-miner/main/source/root/usr/lib/logdata-anomaly-miner/metadata.py'
        http = urllib3.PoolManager()
        r = http.request('GET', url, preload_content=True)
        metadata = r.data.decode()
        http.clear()
        lines = metadata.split('\n')
        curr_version = None
        for line in lines:
            if '__version__ = ' in line:
                curr_version = line.split('__version__ = ')[1].strip('"')
                break
        if __version__ == curr_version:
            print("The current aminer version %s is installed." % curr_version)
        else:
            print("A new aminer version exists (%s). Currently version %s is installed." % (curr_version, __version__))
            print("Use git pull to update the aminer version.")
        sys.exit(0)

    config_file_name = args.config
    run_in_foreground_flag = args.daemon
    run_analysis_child_flag = args.run_analysis
    clear_persistence_flag = args.clear
    remove_persistence_dirs = args.remove
    from_begin_flag = args.from_begin
    global offline_mode  # skipcq: PYL-W0603
    offline_mode = args.offline_mode
    if args.restore is not None and ('.' in args.restore or '/' in args.restore):
        parser.error('The restore path %s must not contain any . or /' % args.restore)
    if args.remove is not None:
        for remove in args.remove:
            if '.' in remove or '/' in remove:
                parser.error('The remove path %s must not contain any . or /' % remove)
    restore_relative_persistence_path = args.restore
    stat_level = 1
    debug_level = 1
    stat_level_console_flag = False
    debug_level_console_flag = False
    if args.stat is not None:
        stat_level = int(args.stat)
        stat_level_console_flag = True
    if args.debug is not None:
        debug_level = int(args.debug)
        debug_level_console_flag = True

    # Load the main configuration file.
    if not os.path.exists(config_file_name):
        print('%s: config "%s" not (yet) available!' % (program_name, config_file_name), file=sys.stderr)
        sys.exit(1)

    # using the solution here to override config_properties:
    # https://stackoverflow.com/questions/27146262/create-variable-key-value-pairs-with-argparse-python
    use_temp_config = False
    config_properties = parse_vars(args.config_properties)
    if args.config_properties and "LearnMode" in config_properties:
        ymlext = [".YAML", ".YML", ".yaml", ".yml"]
        extension = os.path.splitext(config_file_name)[1]
        if extension in ymlext:
            use_temp_config = True
            fd, temp_config = tempfile.mkstemp(suffix=".yml")
            with open(config_file_name) as f:
                for line in f:
                    if "LearnMode" in line:
                        line = "LearnMode: %s" % config_properties["LearnMode"]
                    os.write(fd, line.encode())
            config_file_name = temp_config
            os.close(fd)
        else:
            msg = "The LearnMode parameter does not exist in .py configs!"
            print(msg, sys.stderr)
            logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)
            sys.exit(1)

    # Minimal import to avoid loading too much within the privileged process.
    try:
        aminer_config = AminerConfig.load_config(config_file_name)
        if use_temp_config:
            os.remove(config_file_name)
            config_file_name = args.config
    except ValueError:
        sys.exit(1)

    for config_property in config_properties:
        if config_property == "LearnMode":
            continue
        old_value = aminer_config.config_properties.get(config_property)
        value = config_properties[config_property]
        if old_value is not None:
            try:
                if isinstance(old_value, bool):
                    if value == "True":
                        value = True
                    elif value == "False":
                        value = False
                    else:
                        msg = "The %s parameter must be of type %s!" % (config_property, type(old_value))
                        print(msg, sys.stderr)
                        logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)
                        sys.exit(1)
                elif isinstance(old_value, int):
                    value = int(value)
                elif isinstance(old_value, float):
                    value = float(value)
                elif isinstance(old_value, list):
                    value = ast.literal_eval(value)
            except ValueError:
                msg = "The %s parameter must be of type %s!" % (config_property, type(old_value))
                print(msg, sys.stderr)
                logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)
                sys.exit(1)
        else:
            msg = "The %s parameter is not set in the config. It will be treated as a string!" % config_property
            print("WARNING: " + msg, sys.stderr)
            logging.getLogger(AminerConfig.DEBUG_LOG_NAME).warning(msg)
        aminer_config.config_properties[config_property] = value

    persistence_dir = aminer_config.config_properties.get(AminerConfig.KEY_PERSISTENCE_DIR, AminerConfig.DEFAULT_PERSISTENCE_DIR)
    child_user_name = aminer_config.config_properties.get(AminerConfig.KEY_AMINER_USER)
    child_group_name = aminer_config.config_properties.get(AminerConfig.KEY_AMINER_GROUP)
    child_user_id = -1
    child_group_id = -1
    try:
        if child_user_name is not None:
            child_user_id = getpwnam(child_user_name).pw_uid
        if child_group_name is not None:
            child_group_id = getgrnam(child_group_name).gr_gid
    except:  # skipcq: FLK-E722
        print('Failed to resolve %s or %s' % (AminerConfig.KEY_AMINER_USER, AminerConfig.KEY_AMINER_GROUP), file=sys.stderr)
        sys.exit(1)

    if not stat_level_console_flag and AminerConfig.KEY_LOG_STAT_LEVEL in aminer_config.config_properties:
        stat_level = aminer_config.config_properties[AminerConfig.KEY_LOG_STAT_LEVEL]
    if not debug_level_console_flag and AminerConfig.KEY_LOG_DEBUG_LEVEL in aminer_config.config_properties:
        debug_level = aminer_config.config_properties[AminerConfig.KEY_LOG_DEBUG_LEVEL]
    if AminerConfig.CONFIG_KEY_ENCODING in aminer_config.config_properties:
        AminerConfig.ENCODING = aminer_config.config_properties[AminerConfig.CONFIG_KEY_ENCODING]

    AminerConfig.STAT_LEVEL = stat_level
    AminerConfig.DEBUG_LEVEL = debug_level

    initialize_loggers(aminer_config, child_user_id, child_group_id)

    if restore_relative_persistence_path is not None and (clear_persistence_flag or remove_persistence_dirs):
        msg = 'The --restore parameter removes all persistence files. Do not use this parameter with --Clear or --Remove!'
        print(msg, sys.stderr)
        logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)
        sys.exit(1)

    if clear_persistence_flag:
        if remove_persistence_dirs:
            msg = 'The --clear and --remove arguments must not be used together!'
            print(msg, file=sys.stderr)
            logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)
            sys.exit(1)
        clear_persistence(persistence_dir)

    if remove_persistence_dirs:
        persistence_dir_name = aminer_config.config_properties.get(AminerConfig.KEY_PERSISTENCE_DIR, AminerConfig.DEFAULT_PERSISTENCE_DIR)
        for filename in os.listdir(persistence_dir_name):
            file_path = os.path.join(persistence_dir_name, filename)
            try:
                if not os.path.isdir(file_path):
                    msg = 'The aminer persistence directory should not contain any files.'
                    print(msg, file=sys.stderr)
                    logging.getLogger(AminerConfig.DEBUG_LOG_NAME).warning(msg)
                    continue
                shutil.rmtree(file_path)
            except OSError as e:
                msg = 'Failed to delete %s. Reason: %s' % (file_path, e)
                print(msg, file=sys.stderr)
                logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)

        for filename in remove_persistence_dirs:
            file_path = os.path.join(persistence_dir, filename)
            try:
                if not os.path.exists(file_path):
                    continue
                if not os.path.isdir(file_path):
                    msg = 'The aminer persistence directory should not contain any files.'
                    print(msg, file=sys.stderr)
                    logging.getLogger(AminerConfig.DEBUG_LOG_NAME).warning(msg)
                    continue
                shutil.rmtree(file_path)
            except OSError as e:
                msg = 'Failed to delete %s. Reason: %s' % (file_path, e)
                print(msg, file=sys.stderr)
                logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)

    if restore_relative_persistence_path is not None:
        absolute_persistence_path = os.path.join(persistence_dir, 'backup', restore_relative_persistence_path)
        if not os.path.exists(absolute_persistence_path):
            msg = '%s does not exist. Continuing without restoring persistence.' % absolute_persistence_path
            print(msg, file=sys.stderr)
            logging.getLogger(AminerConfig.DEBUG_LOG_NAME).warning(msg)
        else:
            clear_persistence(persistence_dir)
            copytree(absolute_persistence_path, persistence_dir)
            persistence_dir_fd = SecureOSFunctions.secure_open_base_directory(persistence_dir, os.O_RDONLY | os.O_DIRECTORY | os.O_PATH)
            for dirpath, _dirnames, filenames in os.walk(persistence_dir):
                os.chown(dirpath, child_user_id, child_group_id, dir_fd=persistence_dir_fd, follow_symlinks=False)
                for filename in filenames:
                    os.chown(os.path.join(dirpath, filename), child_user_id, child_user_id, dir_fd=persistence_dir_fd,
                             follow_symlinks=False)

    if from_begin_flag:
        repositioning_data_path = os.path.join(aminer_config.config_properties.get(
            AminerConfig.KEY_PERSISTENCE_DIR, AminerConfig.DEFAULT_PERSISTENCE_DIR), 'AnalysisChild', 'RepositioningData')
        if os.path.exists(repositioning_data_path):
            os.remove(repositioning_data_path)

    if run_analysis_child_flag:
        # Call analysis process, this function will never return.
        run_analysis_child(aminer_config, program_name)

    # Start importing of aminer specific components after reading of "config.py" to allow replacement of components via sys.path
    # from within configuration.
    log_sources_list = aminer_config.config_properties.get(AminerConfig.KEY_LOG_SOURCES_LIST)
    if (log_sources_list is None) or not log_sources_list:
        msg = '%s: %s not defined' % (program_name, AminerConfig.KEY_LOG_SOURCES_LIST)
        print(msg, file=sys.stderr)
        logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)
        sys.exit(1)

    # Now create the management entries for each logfile.
    log_data_resource_dict = {}
    for log_resource_name in log_sources_list:
        # From here on log_resource_name is a byte array.
        log_resource_name = decode_string_as_byte_string(log_resource_name)
        log_resource = None
        if log_resource_name.startswith(b'file://'):
            log_resource = FileLogDataResource(log_resource_name, -1)
        elif log_resource_name.startswith(b'unix://'):
            log_resource = UnixSocketLogDataResource(log_resource_name, -1)
        else:
            msg = 'Unsupported schema in %s: %s' % (AminerConfig.KEY_LOG_SOURCES_LIST, repr(log_resource_name))
            print(msg, file=sys.stderr)
            logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)
            sys.exit(1)
        if not os.path.exists(log_resource_name[7:].decode()):
            msg = "WARNING: file or socket '%s' does not exist (yet)!" % log_resource_name[7:].decode()
            print(msg, file=sys.stderr)
            logging.getLogger(AminerConfig.DEBUG_LOG_NAME).warning(msg)
        try:
            log_resource.open()
        except OSError as open_os_error:
            if open_os_error.errno == errno.EACCES:
                msg = '%s: no permission to access %s' % (program_name, repr(log_resource_name))
                print(msg, file=sys.stderr)
                logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)
                sys.exit(1)
            else:
                msg = '%s: unexpected error opening %s: %d (%s)' % (
                    program_name, repr(log_resource_name), open_os_error.errno, os.strerror(open_os_error.errno))
                print(msg, file=sys.stderr)
                logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)
                sys.exit(1)
        log_data_resource_dict[log_resource_name] = log_resource

    # Create the remote control socket, if any. Do this in privileged mode to allow binding it at arbitrary locations and support restricted
    # permissions of any type for current (privileged) uid.
    remote_control_socket_name = aminer_config.config_properties.get(AminerConfig.KEY_REMOTE_CONTROL_SOCKET_PATH, None)
    remote_control_socket = None
    if remote_control_socket_name is not None:
        if os.path.exists(remote_control_socket_name):
            try:
                os.unlink(remote_control_socket_name)
            except OSError:
                msg = 'Failed to clean up old remote control socket at %s' % remote_control_socket_name
                print(msg, file=sys.stderr)
                logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)
                sys.exit(1)
        # Create the local socket: there is no easy way to create it with correct permissions, hence a fork is needed, setting umask,
        # bind the socket. It is also recommended to create the socket in a directory having the correct permissions already.
        remote_control_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        remote_control_socket.setblocking(False)
        bind_child_pid = os.fork()
        if bind_child_pid == 0:
            os.umask(0o177)
            remote_control_socket.bind(remote_control_socket_name)
            # Do not perform any cleanup, flushing of streams. Use _exit(0) to avoid interference with fork.
            os._exit(0)  # skipcq: PYL-W0212
        os.waitpid(bind_child_pid, 0)
        remote_control_socket.listen(4)

    # Now have checked all we can get from the configuration in the privileged process. Detach from the TTY when in daemon mode.
    if not run_in_foreground_flag:
        child_pid = 0
        try:
            # Fork a child to make sure, we are not the process group leader already.
            child_pid = os.fork()
        except Exception as fork_exception:  # skipcq: PYL-W0703
            msg = 'Failed to daemonize: %s' % fork_exception
            print(msg, file=sys.stderr)
            logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)
            sys.exit(1)
        if child_pid != 0:
            # This is the parent.
            os._exit(0)  # skipcq: PYL-W0212
        # This is the child. Create a new session and become process group leader. Here we get rid of the controlling tty.
        os.setsid()
        # Fork again to become an orphaned process not being session leader, hence not able to get a controlling tty again.
        try:
            child_pid = os.fork()
        except Exception as fork_exception:  # skipcq: PYL-W0703
            msg = 'Failed to daemonize: %s' % fork_exception
            print(msg, file=sys.stderr)
            logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)
            sys.exit(1)
        if child_pid != 0:
            # This is the parent.
            os._exit(0)  # skipcq: PYL-W0212
        # Move to root directory to avoid lingering in some cwd someone else might want to unmount.
        os.chdir('/')
        # Change the umask here to clean all group/other mask bits so that accidentially created files are not accessible by other.
        os.umask(0o77)

    # Install a signal handler catching common stop signals and relaying it to all children for sure.
    # skipcq: PYL-W0603
    global child_termination_triggered_flag
    child_termination_triggered_flag = False

    def graceful_shutdown_handler(_signo, _stackFrame):
        """React on typical shutdown signals."""
        msg = '%s: caught signal, shutting down' % program_name
        print(msg, file=sys.stderr)
        logging.getLogger(AminerConfig.DEBUG_LOG_NAME).info(msg)
        # Just set the flag. It is likely, that child received same signal also so avoid multiple signaling, which could interrupt the
        # shutdown procedure again.
        # skipcq: PYL-W0603
        global child_termination_triggered_flag
        child_termination_triggered_flag = True

    import signal
    signal.signal(signal.SIGHUP, graceful_shutdown_handler)
    signal.signal(signal.SIGINT, graceful_shutdown_handler)
    signal.signal(signal.SIGTERM, graceful_shutdown_handler)

    # Now create the socket to connect the analysis child.
    (parent_socket, child_socket) = socket.socketpair(socket.AF_UNIX, socket.SOCK_DGRAM, 0)
    # Have it nonblocking from here on.
    parent_socket.setblocking(False)
    child_socket.setblocking(False)

    # Use normal fork, we should have been detached from TTY already. Flush stderr to avoid duplication of output if both child and
    # parent want to write something.
    sys.stderr.flush()
    child_pid = os.fork()
    if child_pid == 0:
        # Relocate the child socket fd to 3 if needed
        if child_socket.fileno() != 3:
            os.dup2(child_socket.fileno(), 3)
            child_socket.close()

        # Clear the supplementary groups before dropping privileges. This makes only sense when changing the uid or gid.
        if os.getuid() == 0:
            if ((child_user_id != -1) and (child_user_id != os.getuid())) or ((child_group_id != -1) and (child_group_id != os.getgid())):
                os.setgroups([])

            # Drop privileges before executing child. setuid/gid will raise an exception when call has failed.
            if child_group_id != -1:
                os.setgid(child_group_id)
            if child_user_id != -1:
                os.setuid(child_user_id)
        else:
            msg = 'INFO: No privilege separation when started as unprivileged user'
            print(msg, file=sys.stderr)
            tmp_username = aminer_config.config_properties.get(AminerConfig.KEY_AMINER_USER)
            tmp_group = aminer_config.config_properties.get(AminerConfig.KEY_AMINER_GROUP)
            aminer_user_id = -1
            aminer_group_id = -1
            try:
                if tmp_username is not None:
                    aminer_user_id = getpwnam(tmp_username).pw_uid
                if tmp_group is not None:
                    aminer_group_id = getgrnam(tmp_group).gr_gid
            except:  # skipcq: FLK-E722
                print('Failed to resolve %s or %s' % (AminerConfig.KEY_AMINER_USER, AminerConfig.KEY_AMINER_GROUP), file=sys.stderr)
                sys.exit(1)

            initialize_loggers(aminer_config, aminer_user_id, aminer_group_id)
            logging.getLogger(AminerConfig.DEBUG_LOG_NAME).info(msg)

        # Now resolve the specific analysis configuration file (if any).
        analysis_config_file_name = aminer_config.config_properties.get(AminerConfig.KEY_ANALYSIS_CONFIG_FILE, None)
        if analysis_config_file_name is None:
            analysis_config_file_name = config_file_name
        elif not os.path.isabs(analysis_config_file_name):
            analysis_config_file_name = os.path.join(os.path.dirname(config_file_name), analysis_config_file_name)

        # This is the child. Close all parent file descriptors, we do not need. Perhaps this could be done more elegantly.
        for close_fd in range(4, 1 << 16):
            try:
                os.close(close_fd)
            except OSError as open_os_error:
                if open_os_error.errno == errno.EBADF:
                    continue
                msg = '%s: unexpected exception closing file descriptors: %s' % (program_name, open_os_error)
                print(msg, file=sys.stderr)
                logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)
                # Flush stderr before exit without any cleanup.
                sys.stderr.flush()
                os._exit(1)  # skipcq: PYL-W0212

        # Now execute the very same program again, but user might have moved or renamed it meanwhile. This would be problematic with
        # SUID-binaries (which we do not yet support). Do NOT just fork but also exec to avoid child circumventing
        # parent's ALSR due to cloned kernel VMA.
        exec_args = ['aminerChild', '--run-analysis', '--config', analysis_config_file_name, '--stat', str(stat_level), '--debug',
                     str(debug_level)]
        if offline_mode:
            exec_args.append("--offline-mode")
        if args.config_properties:
            exec_args.append("--config-properties")
            for config_property in args.config_properties:
                exec_args.append(config_property)
        os.execv(sys.argv[0], exec_args)  # skipcq: BAN-B606
        msg = 'Failed to execute child process'
        print(msg, file=sys.stderr)
        logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)
        sys.stderr.flush()
        os._exit(1)  # skipcq: PYL-W0212
    child_socket.close()

    # Send all log resource information currently available to child process.
    for log_resource_name, log_resource in log_data_resource_dict.items():
        if (log_resource is not None) and (log_resource.get_file_descriptor() >= 0):
            SecureOSFunctions.send_logstream_descriptor(parent_socket, log_resource.get_file_descriptor(), log_resource_name)
            log_resource.close()

    # Send the remote control server socket, if any and close it afterwards. It is not needed any more on parent side.
    if remote_control_socket is not None:
        SecureOSFunctions.send_annotated_file_descriptor(parent_socket, remote_control_socket.fileno(), 'remotecontrol', '')
        remote_control_socket.close()

    exit_status = 0
    child_termination_triggered_count = 0
    while True:
        if child_termination_triggered_flag:
            if child_termination_triggered_count == 0:
                time.sleep(1)
            elif child_termination_triggered_count < 5:
                os.kill(child_pid, signal.SIGTERM)
            else:
                os.kill(0, signal.SIGKILL)
            child_termination_triggered_count += 1
        (sig_child_pid, sig_status) = os.waitpid(-1, os.WNOHANG)
        if sig_child_pid != 0:
            if sig_child_pid == child_pid:
                if child_termination_triggered_flag or offline_mode:
                    # This was expected, just terminate.
                    break
                msg = '%s: Analysis child process %d terminated unexpectedly with signal 0x%x' % (program_name, sig_child_pid, sig_status)
                print(msg, file=sys.stderr)
                logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)
                exit_status = 1
                break
            # So the child has been cloned, the clone has terminated. This should not happen either.
            msg = '%s: untracked child %d terminated with with signal 0x%x' % (program_name, sig_child_pid, sig_status)
            print(msg, file=sys.stderr)
            logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)
            exit_status = 1

        # Child information handled, scan for rotated logfiles or other resources, where reopening might make sense.
        for log_resouce_name, log_data_resource in log_data_resource_dict.items():
            try:
                if not log_data_resource.open(reopen_flag=True):
                    continue
            except OSError as open_os_error:
                if open_os_error.errno == errno.EACCES:
                    msg = '%s: no permission to access %s' % (program_name, log_resouce_name)
                    print(msg, file=sys.stderr)
                    logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)
                else:
                    msg = '%s: unexpected error reopening %s: %d (%s)' % (
                        program_name, log_resouce_name, open_os_error.errno, os.strerror(open_os_error.errno))
                    print(msg, file=sys.stderr)
                    logging.getLogger(AminerConfig.DEBUG_LOG_NAME).error(msg)
                exit_status = 2
                continue
            SecureOSFunctions.send_logstream_descriptor(parent_socket, log_data_resource.get_file_descriptor(), log_resouce_name)
            log_data_resource.close()
        time.sleep(1)
    parent_socket.close()
    SecureOSFunctions.close_base_directory()
    SecureOSFunctions.close_log_directory()
    sys.exit(exit_status)


main()



import unittest
from _io import BytesIO
from aminer.parsing.MatchContext import MatchContext
from aminer.parsing.DecimalIntegerValueModelElement import DecimalIntegerValueModelElement
from aminer.analysis.MatchValueStreamWriter import MatchValueStreamWriter
from aminer.parsing.ParserMatch import ParserMatch
from aminer.input.LogAtom import LogAtom
from aminer.parsing.FixedDataModelElement import FixedDataModelElement
from aminer.parsing.SequenceModelElement import SequenceModelElement
from unit.TestBase import TestBase


class MatchValueStreamWriterTest(TestBase):
    """Unittests for the MatchValueStreamWriter."""

    euro = b'Euro '
    match_sequence_s1 = 'match/sequence/s1'
    match_sequence_d1 = 'match/sequence/d1'

    def test1all_atoms_match(self):
        """This test case sets up a set of values, which are all expected to be matched."""
        description = "Test1MatchValueStreamWriter"
        output_stream = BytesIO()
        match_context = MatchContext(b'25537Euro 25538Euro 25539Euro 25540Euro ')
        decimal_integer_value_me = DecimalIntegerValueModelElement('d1', DecimalIntegerValueModelElement.SIGN_TYPE_NONE,
                                                                   DecimalIntegerValueModelElement.PAD_TYPE_NONE)
        fixed_dme = FixedDataModelElement('s1', self.euro)
        sequence_model_element = SequenceModelElement('sequence', [decimal_integer_value_me, fixed_dme])
        match_value_stream_writer = MatchValueStreamWriter(output_stream, [self.match_sequence_d1, self.match_sequence_s1], b';', b'-')
        self.analysis_context.register_component(match_value_stream_writer, description)

        match_element = sequence_model_element.get_match_element('match', match_context)
        log_atom = LogAtom(match_context.match_data, ParserMatch(match_element), 1, match_value_stream_writer)
        match_value_stream_writer.receive_atom(log_atom)

        match_element = sequence_model_element.get_match_element('match', match_context)
        log_atom = LogAtom(match_context.match_data, ParserMatch(match_element), 1, match_value_stream_writer)
        match_value_stream_writer.receive_atom(log_atom)

        match_element = sequence_model_element.get_match_element('match', match_context)
        log_atom = LogAtom(match_context.match_data, ParserMatch(match_element), 1, match_value_stream_writer)
        match_value_stream_writer.receive_atom(log_atom)

        match_element = sequence_model_element.get_match_element('match', match_context)
        log_atom = LogAtom(match_context.match_data, ParserMatch(match_element), 1, match_value_stream_writer)
        match_value_stream_writer.receive_atom(log_atom)

        self.assertEqual(output_stream.getvalue().decode(), '25537;Euro \n25538;Euro \n25539;Euro \n25540;Euro \n')

    def test2all_atoms_match_no_seperator(self):
        """
        This test case sets up a set of values, which are all expected to be matched.
        The seperator string is None, so all values are expected to be one string.
        """
        description = "Test2MatchValueStreamWriter"
        output_stream = BytesIO()
        match_context = MatchContext(b'25537Euro 25538Euro 25539Euro 25540Euro ')
        decimal_integer_value_me = DecimalIntegerValueModelElement('d1', DecimalIntegerValueModelElement.SIGN_TYPE_NONE,
                                                                   DecimalIntegerValueModelElement.PAD_TYPE_NONE)

        fixed_dme = FixedDataModelElement('s1', self.euro)
        sequence_model_element = SequenceModelElement('sequence', [decimal_integer_value_me, fixed_dme])
        match_value_stream_writer = MatchValueStreamWriter(output_stream, [self.match_sequence_d1, self.match_sequence_s1], b'', b'-')
        self.analysis_context.register_component(match_value_stream_writer, description)

        match_element = sequence_model_element.get_match_element('match', match_context)
        log_atom = LogAtom(match_context.match_data, ParserMatch(match_element), 1, match_value_stream_writer)
        match_value_stream_writer.receive_atom(log_atom)

        match_element = sequence_model_element.get_match_element('match', match_context)
        log_atom = LogAtom(match_context.match_data, ParserMatch(match_element), 1, match_value_stream_writer)
        match_value_stream_writer.receive_atom(log_atom)

        match_element = sequence_model_element.get_match_element('match', match_context)
        log_atom = LogAtom(match_context.match_data, ParserMatch(match_element), 1, match_value_stream_writer)
        match_value_stream_writer.receive_atom(log_atom)

        match_element = sequence_model_element.get_match_element('match', match_context)
        log_atom = LogAtom(match_context.match_data, ParserMatch(match_element), 1, match_value_stream_writer)
        match_value_stream_writer.receive_atom(log_atom)

        self.assertEqual(output_stream.getvalue().decode(), '25537Euro \n25538Euro \n25539Euro \n25540Euro \n')

    def test3atom_no_match_missing_value_string_empty(self):
        """
        This test case sets up a set of values, which are all expected to be matched.
        The missing value string is None, so when a string does not match it is simply ignored.
        """
        description = "Test3MatchValueStreamWriter"
        output_stream = BytesIO()
        match_context = MatchContext(b'25537Euro 25538Euro 25539Euro 25540Pfund ')
        decimal_integer_value_me = DecimalIntegerValueModelElement('d1', DecimalIntegerValueModelElement.SIGN_TYPE_NONE,
                                                                   DecimalIntegerValueModelElement.PAD_TYPE_NONE)

        fixed_dme = FixedDataModelElement('s1', self.euro)
        sequence_model_element = SequenceModelElement('sequence', [decimal_integer_value_me, fixed_dme])
        match_value_stream_writer = MatchValueStreamWriter(output_stream, [self.match_sequence_d1, self.match_sequence_s1], b';', b'')
        self.analysis_context.register_component(match_value_stream_writer, description)

        match_element = sequence_model_element.get_match_element('match', match_context)
        log_atom = LogAtom(match_context.match_data, ParserMatch(match_element), 1, match_value_stream_writer)
        match_value_stream_writer.receive_atom(log_atom)

        match_element = sequence_model_element.get_match_element('match', match_context)
        log_atom = LogAtom(match_context.match_data, ParserMatch(match_element), 1, match_value_stream_writer)
        match_value_stream_writer.receive_atom(log_atom)

        match_element = sequence_model_element.get_match_element('match', match_context)
        log_atom = LogAtom(match_context.match_data, ParserMatch(match_element), 1, match_value_stream_writer)
        match_value_stream_writer.receive_atom(log_atom)

        match_element = decimal_integer_value_me.get_match_element('match', match_context)
        match_element.path = self.match_sequence_d1
        log_atom = LogAtom(match_context.match_data, ParserMatch(match_element), 1, match_value_stream_writer)
        match_value_stream_writer.receive_atom(log_atom)

        self.assertEqual(output_stream.getvalue().decode(), '25537;Euro \n25538;Euro \n25539;Euro \n25540;\n')

    def test4atom_no_match_missing_value_string_set(self):
        """
        This test case sets up a set of values, which are all expected to be matched.
        The missing value string is set to a value, so when a string does not match this value is used instead.
        """
        description = "Test4MatchValueStreamWriter"
        output_stream = BytesIO()
        match_context = MatchContext(b'25537Euro 25538Euro 25539Euro 25540Pfund ')
        decimal_integer_value_me = DecimalIntegerValueModelElement('d1', DecimalIntegerValueModelElement.SIGN_TYPE_NONE,
                                                                   DecimalIntegerValueModelElement.PAD_TYPE_NONE)

        fixed_dme = FixedDataModelElement('s1', self.euro)
        sequence_model_element = SequenceModelElement('sequence', [decimal_integer_value_me, fixed_dme])
        match_value_stream_writer = MatchValueStreamWriter(output_stream, [self.match_sequence_d1, self.match_sequence_s1], b';', b'-')
        self.analysis_context.register_component(match_value_stream_writer, description)

        match_element = sequence_model_element.get_match_element('match', match_context)
        log_atom = LogAtom(match_context.match_data, ParserMatch(match_element), 1, match_value_stream_writer)
        match_value_stream_writer.receive_atom(log_atom)

        match_element = sequence_model_element.get_match_element('match', match_context)
        log_atom = LogAtom(match_context.match_data, ParserMatch(match_element), 1, match_value_stream_writer)
        match_value_stream_writer.receive_atom(log_atom)

        match_element = sequence_model_element.get_match_element('match', match_context)
        log_atom = LogAtom(match_context.match_data, ParserMatch(match_element), 1, match_value_stream_writer)
        match_value_stream_writer.receive_atom(log_atom)

        match_element = decimal_integer_value_me.get_match_element('match', match_context)
        match_element.path = self.match_sequence_d1
        log_atom = LogAtom(match_context.match_data, ParserMatch(match_element), 1, match_value_stream_writer)
        match_value_stream_writer.receive_atom(log_atom)

        self.assertEqual(output_stream.getvalue().decode(), '25537;Euro \n25538;Euro \n25539;Euro \n25540;-\n')


if __name__ == "__main__":
    unittest.main()



"""
This module defines a model element that consists of a sequence of model elements that all have to match.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""
from typing import List
from aminer.parsing.MatchElement import MatchElement
from aminer.parsing.ModelElementInterface import ModelElementInterface


class SequenceModelElement(ModelElementInterface):
    """This class defines an element to find matches that comprise matches of all given child model elements."""

    def __init__(self, element_id: str, children: List["ModelElementInterface"]):
        """
        Initialize the ModelElement.
        @param element_id an identifier for the ModelElement which is shown in the path.
        @param children a list of child elements to be iterated through.
        """
        super().__init__(element_id, children=children)

    def get_match_element(self, path, match_context):
        """
        Try to find a match on given data for this model element and all its children.
        When a match is found, the matchContext is updated accordingly.
        @param path the model path to the parent model element invoking this method.
        @param match_context an instance of MatchContext class holding the data context to match against.
        @return the matchElement or None if model did not match.
        """
        current_path = f"{path}/{self.element_id}"
        start_data = match_context.match_data
        matches = []
        for child_element in self.children:
            child_match = child_element.get_match_element(current_path, match_context)
            if child_match is None:
                match_context.match_data = start_data
                return None
            matches += [child_match]

        return MatchElement(current_path, start_data[:len(start_data) - len(match_context.match_data)],
                            start_data[:len(start_data) - len(match_context.match_data)], matches)



import unittest
from aminer.input.JsonStateMachine import json_machine, constant_machine, string_machine, utf8_machine, hex_machine, number_machine,\
    array_machine, object_machine
from unit.TestBase import TestBase


class ByteStreamLineAtomizerTest(TestBase):
    """Unittests for the JsonStateMachine."""

    def test1hex_machine_valid_values(self):
        """Test the hex_machine with all valid four digit values from 0x0000 to 0xFFFF."""
        def check_value(data):  # skipcq: PY-D0003
            self.assertEqual(data, i)
        for i in range(65536):
            string = str(format(i, '#06x')).encode()[2:]  # remove 0x
            state = hex_machine(check_value)
            for c in string:
                state = state(c)
            self.assertIsNone(state)

        for i in range(65536):
            string = str(format(i, '#06x')).upper().encode()[2:]  # remove 0x
            state = hex_machine(check_value)
            for c in string:
                state = state(c)
            self.assertIsNone(state)

    def test2hex_machine_too_short_value(self):
        """Test the hex_machine with too short hex values."""
        def check_value(data):  # skipcq: PY-D0003
            self.assertEqual(data, i)
        for i in range(4096):
            # converts the integer to the shortest possible hex string.
            string = str(hex(i)).encode()[2:]  # remove 0x
            state = hex_machine(check_value)
            for c in string:
                state = state(c)

        i = 4096
        string = str(hex(i)).encode()[2:]  # remove 0x
        state = hex_machine(check_value)
        for c in string:
            state = state(c)
        self.assertIsNone(state)

    def test3hex_machine_too_long_value(self):
        """Test the hex_machine with too long hex values. All values longer than 4 digits are stripped."""
        def check_value(data):  # skipcq: PY-D0003
            self.assertEqual(data, i)
        # only 00FF is read.
        i = 255
        string = b'0x00FFFF'[2:]  # remove 0x
        state = hex_machine(check_value)
        j = 0
        for j, c in enumerate(string):
            state = state(c)
            if state is None:
                break
        self.assertEqual(j, 3)
        self.assertIsNone(state)

        # only 0F12 is read.
        i = 3858
        string = b'0x0F1234'[2:]  # remove 0x
        state = hex_machine(check_value)
        j = 0
        for j, c in enumerate(string):
            state = state(c)
            if state is None:
                break
        self.assertEqual(j, 3)
        self.assertIsNone(state)

    def test4hex_machine_boundary_values(self):
        """Test boundary values before and after 0-9, a-f, A-F."""
        def check_value(data):  # skipcq: PY-D0003
            self.assertEqual(data, i)
        allowed_value_list = '0123456789abcdefABCDEF'
        forbidden_value_list = [int(hex(j), 16) for j in range(48)] + [int(hex(j), 16) for j in range(58, 65)] + [
            int(hex(j), 16) for j in range(71, 97)] + [int(hex(j), 16) for j in range(103, 128)]
        for a in allowed_value_list:
            state = hex_machine(check_value)
            string = '0x'+a+a+a+a
            i = int(string, 16)  # convert hex string to integer
            for _ in range(4):
                state = state(ord(a))
            self.assertEqual(state, None)

        for f in forbidden_value_list:
            state = hex_machine(check_value)
            self.assertIsNone(state(f), "value: %d, char: '%s' should not be allowed in the hex_machine!" % (f, chr(f)))

    def test5hex_machine_started_from_string_machine(self):
        """Test if the hex_machine is started from the string_machine."""
        def check_value(_data):  # skipcq: PY-D0003, PTC-W0049
            pass
        string = b"\u02FF"  # skipcq: PYL-W1402
        state = string_machine(check_value)
        hex_machine_found = False
        for c in string:
            state = state(c)
            if state.__name__ == '_hex':
                hex_machine_found = True
        self.assertIsNone(state(ord(b'"')))
        self.assertTrue(hex_machine_found)

        string = b"\uff02"  # skipcq: PYL-W1402
        state = string_machine(check_value)
        hex_machine_found = False
        for c in string:
            state = state(c)
            if state.__name__ == '_hex':
                hex_machine_found = True
        self.assertIsNone(state(ord(b'"')))
        self.assertTrue(hex_machine_found)

    def test6utf8_machine_allowed_2_byte_values(self):
        """
        Test all allowed values for the utf8_machine with 2 byte values. Only every 4th value is checked to save time.
        This can be changed by changing the step variable. When checking every 4th value the boundary values are also checked.
        """
        def check_value_hex2(data):  # skipcq: PY-D0003
            self.assertEqual(data, (i - 194)*64 + j)
        step = 4
        for i in range(192, 224):
            for j in range(128, 192, step):
                state = utf8_machine(i, check_value_hex2)
                state = state(j)
        # check if the state is None only once to save time.
        self.assertIsNone(state)

    def test7utf8_machine_forbidden_2_byte_boundary_values(self):
        """Test all boundary values for 2 byte utf8 values."""
        def raise_error(_):  # skipcq: PY-D0003
            raise Exception("Valid UTF-8 value found in boundary test!")
        self.assertIsNone(utf8_machine(191, raise_error))
        self.assertIsNone(utf8_machine(192, raise_error)(127))
        self.assertIsNone(utf8_machine(192, raise_error)(192))
        self.assertRaises(Exception, utf8_machine(192, raise_error), 128)
        self.assertRaises(Exception, utf8_machine(192, raise_error), 191)

    def test8utf8_machine_allowed_3_byte_values(self):
        """
        Test all allowed values for the utf8_machine with 3 byte values. Only every 4th value is checked to save time.
        This can be changed by changing the step variable. When checking every 4th value the boundary values are also checked.
        """
        def check_value_hex3(data):  # skipcq: PY-D0003
            self.assertEqual(data, (i - 224)*64*64 + (j - 128)*64 + k - 128)
        step = 4
        for i in range(224, 240):
            for j in range(128, 192, step):
                for k in range(128, 192):
                    state = utf8_machine(i, check_value_hex3)
                    state = state(j)
                    state = state(k)
        # check if the state is None only once to save time.
        self.assertIsNone(state)

    def test9utf8_machine_forbidden_3_byte_boundary_values(self):
        """Test all boundary values for 3 byte utf8 values."""
        def raise_error(_):  # skipcq: PY-D0003
            raise Exception("Valid UTF-8 value found in boundary test!")
        self.assertIsNone(utf8_machine(224, raise_error)(127))
        self.assertIsNone(utf8_machine(224, raise_error)(192))
        self.assertIsNone(utf8_machine(224, raise_error)(128)(127))
        self.assertIsNone(utf8_machine(224, raise_error)(191)(192))
        self.assertRaises(Exception, utf8_machine(224, raise_error)(128), 128)
        self.assertRaises(Exception, utf8_machine(224, raise_error)(191), 191)

    def test10utf8_machine_allowed_4_byte_values(self):
        """
        Test all allowed values for the utf8_machine with 4 byte values. Only every 4th value is checked to save time.
        This can be changed by changing the step variable. When checking every 4th value the boundary values are also checked.
        """
        def check_value_hex4(data):  # skipcq: PY-D0003
            self.assertEqual(data, (i - 240)*64*64*64 + (j - 128)*64*64 + (k - 128)*64 + m - 128)
        step = 4
        for i in range(240, 248):
            for j in range(128, 192, step):
                for k in range(128, 192, step):
                    for m in range(128, 192, step):
                        state = utf8_machine(i, check_value_hex4)
                        state = state(j)
                        state = state(k)
                        state = state(m)
        # check if the state is None only once to save time.
        self.assertIsNone(state)

    def test11utf8_machine_forbidden_3_byte_boundary_values(self):
        """Test all boundary values for 4 byte utf8 values."""
        def raise_error(_):  # skipcq: PY-D0003
            raise Exception("Valid UTF-8 value found in boundary test!")
        self.assertIsNone(utf8_machine(240, raise_error)(127))
        self.assertIsNone(utf8_machine(240, raise_error)(192))
        self.assertIsNone(utf8_machine(240, raise_error)(128)(127))
        self.assertIsNone(utf8_machine(240, raise_error)(191)(192))
        self.assertIsNone(utf8_machine(240, raise_error)(128)(128)(127))
        self.assertIsNone(utf8_machine(240, raise_error)(191)(191)(192))
        self.assertRaises(Exception, utf8_machine(240, raise_error)(128)(128), 128)
        self.assertRaises(Exception, utf8_machine(240, raise_error)(191)(191), 191)

    def test12utf8_machine_started_from_string_machine(self):
        """Test if the utf8_machine is started from the string_machine."""
        def check_value(_data):  # skipcq: PY-D0003, PTC-W0049
            pass
        string = b"File pattern: file\x5f<file-nr>.txt"
        state = string_machine(check_value)
        utf8_machine_found = False
        for c in string:
            state = state(c)
            if state.__name__ == '_utf8':
                utf8_machine_found = True
        self.assertIsNone(state(ord(b'"')))
        self.assertFalse(utf8_machine_found)

        string = b"It is 20\xc2\xb0C"
        state = string_machine(check_value)
        utf8_machine_found = False
        for c in string:
            state = state(c)
            if state.__name__ == '_utf8':
                utf8_machine_found = True
        self.assertIsNone(state(ord(b'"')))
        self.assertTrue(utf8_machine_found)

        string = b"This is a foreign letter: \xe0\xa0\xab"
        state = string_machine(check_value)
        utf8_machine_found = False
        for c in string:
            state = state(c)
            if state.__name__ == '_utf8':
                utf8_machine_found = True
        self.assertIsNone(state(ord(b'"')))
        self.assertTrue(utf8_machine_found)

        string = b"This is an egyptian hieroglyph: \xf0\x93\x80\x90"
        state = string_machine(check_value)
        utf8_machine_found = False
        for c in string:
            state = state(c)
            if state.__name__ == '_utf8':
                utf8_machine_found = True
        self.assertIsNone(state(ord(b'"')))
        self.assertTrue(utf8_machine_found)

    def test13string_machine_valid_values(self):
        """Test the string_machine with all valid characters."""
        def check_value(data):  # skipcq: PY-D0003
            self.assertEqual(data, allowed_chars)
        allowed_chars = "\n"
        for c in range(0x20, 0x80):
            if c in (0x22, 0x5c):  # skip "\
                continue
            allowed_chars += chr(c)
        state = string_machine(check_value)
        for c in allowed_chars.encode():
            state = state(c)
        self.assertEqual(state.__name__, "_string")
        state = state(ord('"'))
        self.assertIsNone(state)

    def test14string_machine_invalid_values(self):
        """Test the string_machine with some invalid values."""
        def raise_error(_):  # skipcq: PY-D0003
            raise Exception("Invalid returned as valid.")
        for c in range(0x20):  # ascii control characters
            if c == 0xa:
                continue
            state = string_machine(raise_error)
            self.assertIsNone(state(c))

        for c in range(0x80, 0xc0):  # some characters after the ascii table
            state = string_machine(raise_error)
            self.assertIsNone(state(c))

    def test15string_machine_escaped_strings(self):
        """Test all allowed escape strings in the string_machine."""
        def check_value(data):  # skipcq: PY-D0003
            self.assertEqual(data, compare_strings)
        escape_strings = b"bf\"\\/"
        compare_strings = "\b\f\"\\/"
        state = string_machine(check_value)
        for c in escape_strings:
            state = state(0x5c)  # \
            state = state(c)
        state = state(0x22)  # "
        self.assertIsNone(state)

    def test16constant_machine_valid_values(self):
        """Test all allowed values for the constant_machine. The first letter was already handled by the json_machine."""
        def check_value(data):  # skipcq: PY-D0003
            self.assertEqual(data, value)
        TRUE = [0x72, 0x75, 0x65]
        FALSE = [0x61, 0x6c, 0x73, 0x65]
        NULL = [0x75, 0x6c, 0x6c]
        value = True
        state = constant_machine(TRUE, True, check_value)
        for t in TRUE:
            state = state(t)
        self.assertIsNone(state)

        value = False
        state = constant_machine(FALSE, False, check_value)
        for f in FALSE:
            state = state(f)
        self.assertIsNone(state)

        value = None
        state = constant_machine(NULL, None, check_value)
        for n in NULL:
            state = state(n)
        self.assertIsNone(state)

    def test17constant_machine_invalid_values(self):
        """Test if constant_machine fails. The first letter was already handled by the json_machine."""
        def raise_error(_):  # skipcq: PY-D0003
            raise Exception("Invalid returned as valid.")
        TRUE = [0x72, 0x75, 0x65]
        TRUE_UPPER = [0x52, 0x55, 0x45]
        FALSE = [0x61, 0x6c, 0x73, 0x65]
        FALSE_UPPER = [0x41, 0x4c, 0x53, 0x45]
        NULL = [0x75, 0x6c, 0x6c]
        NULL_UPPER = [0x55, 0x4c, 0x4c]
        NONE = [0x6f, 0x6e, 0x65]
        state = constant_machine(TRUE, True, raise_error)
        self.assertIsNone(state(TRUE_UPPER[0]))

        state = constant_machine(FALSE, False, raise_error)
        self.assertIsNone(state(FALSE_UPPER[0]))

        state = constant_machine(NULL, None, raise_error)
        self.assertIsNone(state(NULL_UPPER[0]))

        state = constant_machine(NULL, None, raise_error)
        self.assertIsNone(state(NONE[0]))

    def test18constant_machine_started_from_json_machine(self):
        """Test if the constant_machine is started from the json_machine. Due to changes in the json_machine all values must be objects."""
        def check_value(data):  # skipcq: PY-D0003
            self.assertEqual(data, {'var': value})
        OBJECT_PREFIX = [0x7b, 0x22, 0x76, 0x61, 0x72, 0x22, 0x3a, 0x20]  # {"var":
        TRUE = [0x74, 0x72, 0x75, 0x65]
        FALSE = [0x66, 0x61, 0x6c, 0x73, 0x65]
        NULL = [0x6e, 0x75, 0x6c, 0x6c]
        value = True
        state = json_machine(check_value)
        for t in OBJECT_PREFIX + TRUE:
            state = state(t)
        self.assertEqual(state(ord('}')).__name__, '_value')

        value = False
        state = json_machine(check_value)
        for f in OBJECT_PREFIX + FALSE:
            state = state(f)
        self.assertEqual(state(ord('}')).__name__, '_value')

        value = None
        state = json_machine(check_value)
        for n in OBJECT_PREFIX + NULL:
            state = state(n)
        self.assertEqual(state(ord('}')).__name__, '_value')

    def check_number_machine(self, check_int_value, value, end_sign):  # skipcq: PY-D0003
        state = number_machine(value[0], check_int_value)
        for c in value[1:]:
            state = state(c)
        self.assertIsNone(state(end_sign))

    def test19number_machine_valid_values(self):
        """Test valid values in the number_machine."""
        def check_int_value(data, byte_data):  # skipcq: PY-D0003
            self.assertEqual(data, int(value))
            self.assertEqual(end_sign, byte_data)

        def check_float_value(data, byte_data):  # skipcq: PY-D0003
            self.assertEqual(round(data, 10), float(value))
            self.assertEqual(end_sign, byte_data)
        end_sign = ord(',')
        value = b'222'
        self.check_number_machine(check_int_value, value, end_sign)

        value = b'9223372036854775808'  # maxsize 2^64
        self.check_number_machine(check_int_value, value, end_sign)

        value = b'-222'
        self.check_number_machine(check_int_value, value, end_sign)

        value = b'+222'
        self.check_number_machine(check_int_value, value, end_sign)

        value = b'21.50'
        self.check_number_machine(check_float_value, value, end_sign)

        value = b'21.05'
        self.check_number_machine(check_float_value, value, end_sign)

        value = b'-21.05'
        self.check_number_machine(check_float_value, value, end_sign)

        value = b'1.56E-5'
        self.check_number_machine(check_float_value, value, end_sign)

        value = b'1.56e-5'
        self.check_number_machine(check_float_value, value, end_sign)

    def test20number_machine_end_signs(self):
        """Check if all non numerical signs end the number_machine."""
        def check_int_value(data, byte_data):  # skipcq: PY-D0003
            self.assertEqual(data, int(value))
            self.assertEqual(end_sign, byte_data)
        value = b'222'
        end_signs = list(range(0x2e)) + list(range(0x3a, 0x45)) + list(range(0x46, 0x65)) + list(range(0x66, 0x80))
        valid_signs = [0x2e, 0x45, 0x65] + list(range(0x30, 0x39))
        for end_sign in end_signs:
            state = number_machine(value[0], check_int_value)
            for c in value[1:]:
                state = state(c)
            self.assertIsNone(state(end_sign))
        for end_sign in valid_signs:
            state = number_machine(value[0], check_int_value)
            for c in value[1:]:
                state = state(c)
            self.assertIsNotNone(state(end_sign))

    def test21number_machine_invalid_values(self):
        """Test invalid values in the number_machine."""
        def raise_error(_data, _byte_data):  # skipcq: PY-D0003
            raise Exception("Invalid number treated as valid!")
        value = b'- 222'
        state = number_machine(value[0], raise_error)
        self.assertIsNone(state(value[1]))

        # octal number
        value = b'0222'
        self.assertIsNone(number_machine(value[0], raise_error)(value[1]))

        # negative octal number
        value = b'-0222'
        self.assertIsNone(number_machine(value[0], raise_error)(value[1])(value[2]))

        # hex number
        value = b'0x80'
        self.assertIsNone(number_machine(value[0], raise_error)(value[1]))

        value = b'NaN'
        self.assertIsNone(number_machine(value[0], raise_error))

        value = b'Infinity'
        self.assertIsNone(number_machine(value[0], raise_error))

        value = b'.1'
        self.assertIsNone(number_machine(value[0], raise_error))

    def check_number_machine_from_json_machine(self, check_int_value, value, end_sign):  # skipcq: PY-D0003
        state = json_machine(check_int_value)
        for c in value:
            state = state(c)
        self.assertEqual(state(end_sign).__name__, '_value')

    def test22number_machine_started_from_json_machine(self):
        """Test if the number_machine is started from the json_machine."""
        def check_int_value(data):  # skipcq: PY-D0003
            self.assertEqual(data, {'value': int(value)})

        def check_float_value(data):  # skipcq: PY-D0003
            data['value'] = round(data['value'], 10)
            self.assertEqual(data, {'value': float(value)})
        end_sign = ord('}')
        object_prefix = b'{"value": '
        value = b'222'
        self.check_number_machine_from_json_machine(check_int_value, object_prefix+value, end_sign)

        value = b'9223372036854775808'  # maxsize 2^64
        self.check_number_machine_from_json_machine(check_int_value, object_prefix+value, end_sign)

        value = b'-222'
        self.check_number_machine_from_json_machine(check_int_value, object_prefix+value, end_sign)

        value = b'+222'
        self.check_number_machine_from_json_machine(check_int_value, object_prefix+value, end_sign)

        value = b'0'
        self.check_number_machine_from_json_machine(check_float_value, object_prefix+value, end_sign)

        value = b'21.50'
        self.check_number_machine_from_json_machine(check_float_value, object_prefix+value, end_sign)

        value = b'21.05'
        self.check_number_machine_from_json_machine(check_float_value, object_prefix+value, end_sign)

        value = b'-21.05'
        self.check_number_machine_from_json_machine(check_float_value, object_prefix+value, end_sign)

        value = b'0.56'
        self.check_number_machine_from_json_machine(check_float_value, object_prefix+value, end_sign)

        value = b'1.56E-5'
        self.check_number_machine_from_json_machine(check_float_value, object_prefix+value, end_sign)

        value = b'1.56e-5'
        self.check_number_machine_from_json_machine(check_float_value, object_prefix+value, end_sign)

        value = b'1.56e+5'
        self.check_number_machine_from_json_machine(check_float_value, object_prefix+value, end_sign)

        value = b'0.56e+5'
        self.check_number_machine_from_json_machine(check_float_value, object_prefix+value, end_sign)

    def test23array_machine_valid_array(self):
        """Test possible valid arrays."""
        def check_value(data):  # skipcq: PY-D0003
            self.assertEqual(data, compare_value)
        value = b'"string", 22, 22.50, true, false, null]'
        compare_value = ['string', 22, 22.5, True, False, None]
        state = array_machine(check_value)
        for c in value:
            state = state(c)
        self.assertIsNone(state)

        value = b'\n\t\t"string",\n\t\t22,\n\t\t22.50,\n\t\ttrue,\n\t\tfalse,\n\t\tnull]'
        state = array_machine(check_value)
        for c in value:
            state = state(c)
        self.assertIsNone(state)

        value = b'{"value": 22}, {"value": "string"}]'
        compare_value = [{'value': 22}, {'value': 'string'}]
        state = array_machine(check_value)
        for c in value:
            state = state(c)
        self.assertIsNone(state)

    def test24array_machine_invalid_formats(self):
        """Test the array_machine with invalid formats."""
        def raise_error(_):  # skipcq: PY-D0003
            raise Exception("Invalid returned as valid.")
        value = b'"string" 22, 22.50, true, false, null]'
        state = array_machine(raise_error)
        for c in value[:value.index(b'2') + 1]:
            state = state(c)
        self.assertIsNone(state)

        value = b'"key": {"value": 2}]'
        state = array_machine(raise_error)
        for c in value[:value.index(b':') + 1]:
            state = state(c)
        self.assertIsNone(state)

    def test25array_machine_started_from_json_machine(self):
        """Test if the array_machine is started from the json_machine."""
        def check_value(data):  # skipcq: PY-D0003
            self.assertEqual(data, compare_value)
        value = b'{"values_array": ["string", 22, 22.50, true, false, null]}'
        compare_value = {'values_array': ['string', 22, 22.5, True, False, None]}
        state = json_machine(check_value)
        for c in value:
            state = state(c)
        self.assertEqual(state.__name__, '_value')

        value = b'{"values_array": [\n\t\t"string",\n\t\t22,\n\t\t22.50,\n\t\ttrue,\n\t\tfalse,\n\t\tnull]}'
        state = json_machine(check_value)
        for c in value:
            state = state(c)
        self.assertEqual(state.__name__, '_value')

        value = b'{"objects_array": [{"value": 22}, {"value": "string"}]}'
        compare_value = {'objects_array': [{'value': 22}, {'value': 'string'}]}
        state = json_machine(check_value)
        for c in value:
            state = state(c)
        self.assertEqual(state.__name__, '_value')

    def test26object_machine_valid_objects(self):
        """Check if the object_machine can handle different valid formats."""
        def check_value(data):  # skipcq: PY-D0003
            self.assertEqual(data, compare_value)
        # single line, no spaces
        value = b'"string":"Hello World","integer":22,"float":22.23,"bool":true,"array":["Hello","World"]}'
        compare_value = {'string': 'Hello World', 'integer': 22, 'float': 22.23, 'bool': True, 'array': ['Hello', 'World']}
        state = object_machine(check_value)
        for c in value:
            state = state(c)
        self.assertIsNone(state)
        # single line with spaces
        value = b'"string": "Hello World", "integer": 22, "float": 22.23, "bool": true, "array": ["Hello", "World"]}'
        state = object_machine(check_value)
        for c in value:
            state = state(c)
        self.assertIsNone(state)
        # multiline with tabs
        value = b'\n\t"string": "Hello World",\n\t"integer": 22,\n\t"float": 22.23,\n\t"bool": true,\n\t"array": [' \
                b'\n\t\t"Hello",\n\t\t"World"]}'
        state = object_machine(check_value)
        for c in value:
            state = state(c)
        self.assertIsNone(state)

    def test27object_machine_invalid_values(self):
        """Test the object_machine with invalid values."""
        def raise_error(_):  # skipcq: PY-D0003
            raise Exception("Invalid returned as valid.")
        # keys without "
        value = b'"string":"Hello World",integer:22,"float":22.23,"bool":true,"array":["Hello","World"]}'
        state = object_machine(raise_error)
        for c in value[:value.index(b'integer') + 1]:
            state = state(c)
        self.assertIsNone(state)
        # = instead of :
        value = b'"string":"Hello World","integer"=22,"float":22.23,"bool":true,"array":["Hello","World"]}'
        state = object_machine(raise_error)
        for c in value[:value.index(b'=') + 1]:
            state = state(c)
        self.assertIsNone(state)
        # no comma after attribute. The error is only found after the next :. However this behavior is not problematic, because another
        # attribute or the end bracket } has to follow.
        value = b'"string":"Hello World","integer":22 "float":22.23,"bool":true,"array":["Hello","World"]}'
        state = object_machine(raise_error)
        for c in value[:value.index(b':22.') + 1]:
            state = state(c)
        self.assertIsNone(state)

    def test28object_machine_started_from_json_machine(self):
        """Test if the object_machine is started from the json_machine."""
        def check_value(data):  # skipcq: PY-D0003
            self.assertEqual(data, compare_value)
        # single line, no spaces
        value = b'{"string":"Hello World","integer":22,"float":22.23,"bool":true,"array":["Hello","World"]}'
        compare_value = {'string': 'Hello World', 'integer': 22, 'float': 22.23, 'bool': True, 'array': ['Hello', 'World']}
        state = json_machine(check_value)
        for c in value:
            state = state(c)
        self.assertEqual(state.__name__, '_value')
        # single line with spaces
        value = b'{"string": "Hello World", "integer": 22, "float": 22.23, "bool": true, "array": ["Hello", "World"]}'
        state = json_machine(check_value)
        for c in value:
            state = state(c)
        self.assertEqual(state.__name__, '_value')
        # multiline with tabs
        value = b'{\n\t"string": "Hello World",\n\t"integer": 22,\n\t"float": 22.23,\n\t"bool": true,\n\t"array": [' \
                b'\n\t\t"Hello",\n\t\t"World"]}'
        state = json_machine(check_value)
        for c in value:
            state = state(c)
        self.assertEqual(state.__name__, '_value')

        value = b"""{"HistogramData": {"Bins": {"...-0]": 0, "[0-1]": 0, "[1-2]": 0, "[2-3]": 0, "[3-...]": 0\n}, "BinNames": ["...-0]",
"[0-1]", "[1-2]", "[2-3]", "[3-...]"]}}"""
        compare_value = {"HistogramData": {"Bins": {'...-0]': 0, '[0-1]': 0, '[1-2]': 0, '[2-3]': 0, '[3-...]': 0}, "BinNames": [
            '...-0]', '[0-1]', '[1-2]', '[2-3]', '[3-...]']}}
        state = json_machine(check_value)
        for c in value:
            state = state(c)
        self.assertEqual(state.__name__, '_value')

    def test29json_machine_only_allow_objects_at_start(self):
        """The json_machine must only allow objects at the start."""
        def raise_error(_):  # skipcq: PY-D0003
            raise Exception("Invalid returned as valid.")
        forbidden_values = [0x22, 0x2b, 0x2d, 0x31, 0x5b, 0x74, 0x66, 0x6e]
        for value in forbidden_values:
            state = json_machine(raise_error)
            self.assertIsNone(state(value))
        state = json_machine(raise_error)
        self.assertIsNotNone(state(ord('{')))


if __name__ == "__main__":
    unittest.main()



import unittest
from aminer.analysis.EventCorrelationDetector import EventCorrelationDetector, set_random_seed
from aminer.input.LogAtom import LogAtom
from aminer.parsing.MatchContext import MatchContext
from aminer.parsing.ParserMatch import ParserMatch
from aminer.parsing.FirstMatchModelElement import FirstMatchModelElement
from aminer.parsing.FixedDataModelElement import FixedDataModelElement
from unit.TestBase import TestBase
from time import time
import random


class EventCorrelationDetectorTest(TestBase):
    """Unittests for the EventCorrelationDetector."""

    alphabet = b'abcdefghijklmnopqrstuvwxyz'
    alphabet_model = None
    analysis = 'Analysis.%s'

    @classmethod
    def setUpClass(cls):
        """Set up the data for the all tests."""
        children = []
        for _, val in enumerate(cls.alphabet):
            char = bytes([val])
            children.append(FixedDataModelElement(char.decode(), char))
        cls.alphabet_model = FirstMatchModelElement('first', children)
        error_rate = 0.000085
        cls.perfect_data_diff5 = cls.generate_perfect_data(cls, 30000, 5)
        cls.perfect_data_diff1 = cls.generate_perfect_data(cls, 30000, 1)
        cls.errored_data_diff5 = cls.generate_errored_data(cls, 100000, 5, error_rate)
        cls.errored_data_diff1 = cls.generate_errored_data(cls, 100000, 1, error_rate)
        cls.errored_data_diff5_low_error_rate = cls.generate_errored_data(cls, 100000, 5, error_rate / 2.5)
        cls.errored_data_diff1_low_error_rate = cls.generate_errored_data(cls, 100000, 1, error_rate / 2.5)
        set_random_seed(42)

    def test1learn_from_clear_examples(self):
        """In this test case perfect examples are used to learn and evaluate rules. The default parameters are used."""
        description = 'test1eventCorrelationDetectorTest'
        ecd = EventCorrelationDetector(
            self.aminer_config, [self.stream_printer_event_handler], check_rules_flag=True, learn_mode=True)
        self.analysis_context.register_component(ecd, description)
        self.run_ecd_test(ecd, self.perfect_data_diff5[:12000])

        ecd = EventCorrelationDetector(
            self.aminer_config, [self.stream_printer_event_handler], check_rules_flag=True, learn_mode=True)
        self.analysis_context.register_component(ecd, description + '2')
        self.run_ecd_test(ecd, self.perfect_data_diff1[:12000])

    def test2learn_from_clear_examples_with_smaller_probabilities(self):
        """
        Like in test1 perfect examples are used.
        The generation_probability and generation_factor are set to 0.5 in the first case and 0.1 in the second case.
        The EventCorrelationDetector should still learn the rules as expected.
        """
        description = 'test2eventCorrelationDetectorTest'
        ecd = EventCorrelationDetector(self.aminer_config, [self.stream_printer_event_handler], check_rules_flag=True,
                                       generation_probability=0.5, generation_factor=0.5, learn_mode=True)
        self.analysis_context.register_component(ecd, description)
        self.run_ecd_test(ecd, self.perfect_data_diff5[:30000])

        ecd = EventCorrelationDetector(self.aminer_config, [self.stream_printer_event_handler], check_rules_flag=True,
                                       generation_probability=0.5, generation_factor=0.5, learn_mode=True)
        self.analysis_context.register_component(ecd, description + '2')
        self.run_ecd_test(ecd, self.perfect_data_diff1[:30000])

        ecd = EventCorrelationDetector(self.aminer_config, [self.stream_printer_event_handler], check_rules_flag=True,
                                       generation_probability=0.3, generation_factor=0.3, learn_mode=True)
        self.analysis_context.register_component(ecd, description + '3')
        self.run_ecd_test(ecd, self.perfect_data_diff5[:100000])

        ecd = EventCorrelationDetector(self.aminer_config, [self.stream_printer_event_handler], check_rules_flag=True,
                                       generation_probability=0.3, generation_factor=0.3, learn_mode=True)
        self.analysis_context.register_component(ecd, description + '4')
        self.run_ecd_test(ecd, self.perfect_data_diff1[:100000])

    def test3learn_from_examples_with_errors(self):
        """In this test case examples with errors are used, but still should be learned. The same parameters like in test1 are used."""
        description = 'test3eventCorrelationDetectorTest'
        ecd = EventCorrelationDetector(
            self.aminer_config, [self.stream_printer_event_handler], check_rules_flag=True, learn_mode=True)
        self.analysis_context.register_component(ecd, description)
        self.run_ecd_test(ecd, self.errored_data_diff5[:12000])

        ecd = EventCorrelationDetector(
            self.aminer_config, [self.stream_printer_event_handler], check_rules_flag=True, learn_mode=True)
        self.analysis_context.register_component(ecd, description + '2')
        self.run_ecd_test(ecd, self.errored_data_diff1[:12000])

    def test4learn_from_examples_with_errors_and_smaller_probabilities(self):
        """
        In this test case examples with errors are used, but still should be learned.
        These tests are using a higher generation_probability and generation_factor, because the data contains errors.
        """
        description = 'test4eventCorrelationDetectorTest'
        ecd = EventCorrelationDetector(self.aminer_config, [self.stream_printer_event_handler], check_rules_flag=True,
                                       generation_probability=0.7, generation_factor=0.99, learn_mode=True)
        self.analysis_context.register_component(ecd, description)
        self.run_ecd_test(ecd, self.errored_data_diff5_low_error_rate[:25000])

        ecd = EventCorrelationDetector(self.aminer_config, [self.stream_printer_event_handler], check_rules_flag=True,
                                       generation_probability=0.7, generation_factor=0.99, learn_mode=True)
        self.analysis_context.register_component(ecd, description + '2')
        self.run_ecd_test(ecd, self.errored_data_diff1_low_error_rate[:25000])

        ecd = EventCorrelationDetector(self.aminer_config, [self.stream_printer_event_handler], check_rules_flag=True,
                                       generation_probability=0.5, generation_factor=0.95, learn_mode=True)
        self.analysis_context.register_component(ecd, description + '3')
        self.run_ecd_test(ecd, self.errored_data_diff5_low_error_rate[:40000])

        ecd = EventCorrelationDetector(self.aminer_config, [self.stream_printer_event_handler], check_rules_flag=True,
                                       generation_probability=0.5, generation_factor=0.95, learn_mode=True)
        self.analysis_context.register_component(ecd, description + '4')
        self.run_ecd_test(ecd, self.errored_data_diff1_low_error_rate[:40000])

    def test5learn_safe_assumptions(self):
        """
        In this test case p0 and alpha are chosen carefully to only find safe assumptions about the implications in the data.
        Therefor more iterations in the training phase are needed.
        """
        description = 'test5eventCorrelationDetectorTest'
        ecd = EventCorrelationDetector(
            self.aminer_config, [self.stream_printer_event_handler], check_rules_flag=True, p0=1.0, alpha=0.01, learn_mode=True)
        self.analysis_context.register_component(ecd, description)
        self.run_ecd_test(ecd, self.perfect_data_diff5[:20000])

        ecd = EventCorrelationDetector(
            self.aminer_config, [self.stream_printer_event_handler], check_rules_flag=True, p0=1.0, alpha=0.01, learn_mode=True)
        self.analysis_context.register_component(ecd, description + '2')
        self.run_ecd_test(ecd, self.errored_data_diff5_low_error_rate[:40000])

        ecd = EventCorrelationDetector(
            self.aminer_config, [self.stream_printer_event_handler], check_rules_flag=True, p0=1.0, alpha=0.01, learn_mode=True)
        self.analysis_context.register_component(ecd, description + '3')
        self.run_ecd_test(ecd, self.perfect_data_diff1[:20000])

        ecd = EventCorrelationDetector(
            self.aminer_config, [self.stream_printer_event_handler], check_rules_flag=True, p0=1.0, alpha=0.01, learn_mode=True)
        self.analysis_context.register_component(ecd, description + '4')
        self.run_ecd_test(ecd, self.errored_data_diff1_low_error_rate[:40000])

    def test6approximately_learn_implications(self):
        """
        In this unittest p0 and alpha are chosen to approximately find sequences in log data.
        Therefor not as many iterations are needed to learn the rules.
        """
        description = 'test6eventCorrelationDetectorTest'
        ecd = EventCorrelationDetector(
            self.aminer_config, [self.stream_printer_event_handler], check_rules_flag=True, p0=0.7, alpha=0.1, learn_mode=True)
        self.analysis_context.register_component(ecd, description)
        self.run_ecd_test(ecd, self.perfect_data_diff5[:10000])

        ecd = EventCorrelationDetector(
            self.aminer_config, [self.stream_printer_event_handler], check_rules_flag=True, p0=0.7, alpha=0.1, learn_mode=True)
        self.analysis_context.register_component(ecd, description + '2')
        self.run_ecd_test(ecd, self.errored_data_diff5[:10000])

        ecd = EventCorrelationDetector(
            self.aminer_config, [self.stream_printer_event_handler], check_rules_flag=True, p0=0.7, alpha=0.1, learn_mode=True)
        self.analysis_context.register_component(ecd, description + '3')
        self.run_ecd_test(ecd, self.perfect_data_diff1[:10000])

        ecd = EventCorrelationDetector(
            self.aminer_config, [self.stream_printer_event_handler], check_rules_flag=True, p0=0.7, alpha=0.1, learn_mode=True)
        self.analysis_context.register_component(ecd, description + '4')
        self.run_ecd_test(ecd, self.errored_data_diff1[:10000])

    def test7constraint_list(self):
        """Test the allowlisting of paths."""
        description = 'test7eventCorrelationDetectorTest'
        ecd = EventCorrelationDetector(
            self.aminer_config, [self.stream_printer_event_handler], check_rules_flag=True, p0=0.7, alpha=0.1, learn_mode=True)
        self.analysis_context.register_component(ecd, description)
        self.assertEqual([], ecd.constraint_list)

        match_context_fixed_dme = MatchContext(b' pid=')
        fixed_dme = FixedDataModelElement('s1', b' pid=')
        match_element_fixed_dme = fixed_dme.get_match_element("", match_context_fixed_dme)

        # unknown path
        ecd.allowlist_event(self.analysis % ecd.__class__.__name__, match_element_fixed_dme.get_path(), None)
        self.assertEqual(['/s1'], ecd.constraint_list)

        # known path
        ecd.allowlist_event(self.analysis % ecd.__class__.__name__, match_element_fixed_dme.get_path(), None)
        self.assertEqual(['/s1'], ecd.constraint_list)

    def check_rules(self, sorted_back_rules, sorted_forward_rules, diff):
        """Check if the rules are as expected."""
        for path in sorted_forward_rules:
            self.assertEqual(len(sorted_forward_rules[path]), 5 / diff)
            implications = []
            trigger = b''
            for rule in sorted_forward_rules[path]:
                trigger = rule.trigger_event[0].split('/')[-1].encode()
                implications.append(self.alphabet.index(rule.implied_event[0].split('/')[-1].encode()))
            for i in range(1, len(sorted_forward_rules[path]), 1):  # skipcq: PTC-W0060
                self.assertIn((self.alphabet.index(trigger) + i) % len(self.alphabet), implications)
        for path in sorted_back_rules:
            self.assertEqual(len(sorted_back_rules[path]), 5 / diff)
            trigger = b''
            implications = []
            for rule in sorted_back_rules[path]:
                trigger = rule.trigger_event[0].split('/')[-1].encode()
                implications.append(self.alphabet.index(rule.implied_event[0].split('/')[-1].encode()))
            for i in range(1, len(sorted_back_rules[path]), 1):  # skipcq: PTC-W0060
                self.assertIn((self.alphabet.index(trigger) - i) % len(self.alphabet), implications)

    def check_anomaly_detection(self, ecd, t, diff):
        """Check if anomalies were detected as expected."""
        for char in self.alphabet:
            self.reset_output_stream()
            char = bytes([char])
            parser_match = ParserMatch(self.alphabet_model.get_match_element('parser', MatchContext(char)))
            t += 5 * 3
            ecd.receive_atom(LogAtom(char, parser_match, t, self.__class__.__name__))
            # another LogAtom must be received to check the follow anomalies.
            t += 5 * 3
            ecd.receive_atom(LogAtom(char, parser_match, t, self.__class__.__name__))

            # precede anomaly
            for i in range(1, int(5 / diff) + 1, 1):
                self.assertIn('Event %s is missing, but should precede event %s' % (
                    repr(bytes([self.alphabet[(self.alphabet.index(char) - i) % len(self.alphabet)]])), repr(char)),
                    self.output_stream.getvalue())
            for i in range(int(5 / diff) + 1, len(self.alphabet), 1):  # skipcq: PTC-W0060
                self.assertNotIn('Event %s is missing, but should precede event %s' % (
                    repr(bytes([self.alphabet[(self.alphabet.index(char) - i) % len(self.alphabet)]])), repr(char)),
                    self.output_stream.getvalue())

            # follow anomaly
            for i in range(1, int(5 / diff) + 1, 1):
                self.assertIn('Event %s is missing, but should follow event %s' % (
                    repr(bytes([self.alphabet[(self.alphabet.index(char) + i) % len(self.alphabet)]])), repr(char)),
                    self.output_stream.getvalue())
            for i in range(int(5 / diff) + 1, len(self.alphabet), 1):  # skipcq: PTC-W0060
                self.assertNotIn('Event %s is missing, but should follow event %s' % (
                    repr(bytes([self.alphabet[(self.alphabet.index(char) + i) % len(self.alphabet)]])), repr(char)),
                    self.output_stream.getvalue())

    def run_ecd_test(self, ecd, log_atoms):
        """Run the ECD test."""
        diff = log_atoms[1].atom_time - log_atoms[0].atom_time
        log_atom = None
        for log_atom in log_atoms:
            ecd.receive_atom(log_atom)
        sorted_forward_rules = dict(sorted(ecd.forward_rules.items()))
        sorted_back_rules = dict(sorted(ecd.back_rules.items()))
        self.assertEqual(len(sorted_forward_rules), len(self.alphabet_model.children))
        self.assertEqual(len(sorted_back_rules), len(self.alphabet_model.children))
        self.check_rules(sorted_back_rules, sorted_forward_rules, diff)
        ecd.learn_mode = False
        self.check_anomaly_detection(ecd, log_atom.atom_time, diff)

    def generate_perfect_data(self, iterations, diff):
        """Generate data without any error."""
        log_atoms = []
        t = time()
        for i in range(1, iterations+1):
            char = bytes([self.alphabet[i % len(self.alphabet)]])
            parser_match = ParserMatch(self.alphabet_model.get_match_element('parser', MatchContext(char)))
            t += diff
            log_atoms.append(LogAtom(char, parser_match, t, self.__class__.__name__))
        return log_atoms

    def generate_errored_data(self, iterations, diff, error_rate):
        """Generate data with errors according to the error_rate."""
        log_atoms = []
        t = time()
        divisor = 1
        while error_rate * divisor < 1:
            divisor = divisor * 10
        err = divisor * error_rate
        divisor //= err
        for i in range(1, iterations+1):
            if i % divisor == 0 and i != 0:
                char = bytes([self.alphabet[int(i + random.uniform(diff+1, len(self.alphabet))) % len(self.alphabet)]])
            else:
                char = bytes([self.alphabet[i % len(self.alphabet)]])
            parser_match = ParserMatch(self.alphabet_model.get_match_element('parser', MatchContext(char)))
            t += diff
            log_atoms.append(LogAtom(char, parser_match, t, self.__class__.__name__))
        return log_atoms


if __name__ == "__main__":
    unittest.main()



import unittest
from aminer.input.ByteStreamLineAtomizer import ByteStreamLineAtomizer
from aminer.parsing.MatchContext import MatchContext
from aminer.parsing.AnyByteDataModelElement import AnyByteDataModelElement
import sys
from unit.TestBase import TestBase


class ByteStreamLineAtomizerTest(TestBase):
    """Unittests for the ByteStreamLineAtomizer."""

    illegal_access1 = b'WARNING: All illegal access operations will be denied in a future release'
    illegal_access2 = 'WARNING: All illegal access operations will be denied in a future release\n\n'

    def test1normal_line(self):
        """A normal line is tested as Input of the Class."""
        any_dme = AnyByteDataModelElement('s')
        byte_stream_line_atomizer = ByteStreamLineAtomizer(any_dme, [], [self.stream_printer_event_handler], 300, [])
        data = self.illegal_access1 + b'\n'
        self.assertEqual(byte_stream_line_atomizer.consume_data(data, False), len(data))

    def test2normal_complete_overlong_line(self):
        """A complete, overlong line is tested as Input of the Class."""
        match_context_fixed_dme = MatchContext(self.illegal_access1)
        any_dme = AnyByteDataModelElement('s')
        _match_element_fixed_dme = any_dme.get_match_element("match1", match_context_fixed_dme)

        byte_stream_line_atomizer = ByteStreamLineAtomizer(
            any_dme, [], [self.stream_printer_event_handler], sys.getsizeof(match_context_fixed_dme.match_data) - 1, [])
        self.assertGreater(
            byte_stream_line_atomizer.consume_data(self.illegal_access1 + b'\n', False), 0)
        self.assertEqual(self.output_stream.getvalue(), 'Overlong line detected (1 lines)\n  %s' % self.illegal_access2)

    def test3normal_incomplete_overlong_line_stream_not_ended(self):
        """A incomplete, overlong line, with the stream NOT ended, is tested as Input of the Class."""
        match_context_fixed_dme = MatchContext(self.illegal_access1)
        any_dme = AnyByteDataModelElement('s')
        _match_element_fixed_dme = any_dme.get_match_element("match1", match_context_fixed_dme)

        byte_stream_line_atomizer = ByteStreamLineAtomizer(
            any_dme, [], [self.stream_printer_event_handler], sys.getsizeof(match_context_fixed_dme.match_data) - 1, [])
        self.assertGreater(byte_stream_line_atomizer.consume_data(self.illegal_access1, False), 0)
        self.assertEqual(self.output_stream.getvalue(), 'Start of overlong line detected (1 lines)\n  %s' % self.illegal_access2)

    def test4normal_incomplete_overlong_line_stream_ended(self):
        """A incomplete, overlong line, with the stream ended, is tested as Input of the Class."""
        any_dme = AnyByteDataModelElement('s')
        byte_stream_line_atomizer = ByteStreamLineAtomizer(any_dme, [], [self.stream_printer_event_handler], 300, [])
        self.assertGreater(byte_stream_line_atomizer.consume_data(self.illegal_access1, True), 0)
        self.assertEqual(self.output_stream.getvalue(), 'Incomplete last line (1 lines)\n  %s' % self.illegal_access2)

    def test5eol_sep(self):
        """Test the eol_sep parameter."""
        searched = b'WARNING: All illegal access operations will be denied in a future release\nAnother line of data\nThe third line of ' \
                   b'data.'
        any_dme = AnyByteDataModelElement('s')
        byte_stream_line_atomizer = ByteStreamLineAtomizer(any_dme, [], [self.stream_printer_event_handler], 300, [], eol_sep=b'\n\n')
        data = searched + b'\n\nother line'
        self.assertEqual(byte_stream_line_atomizer.consume_data(data, False), len(data)-len(b'other line'))

    def test6json_format(self):
        """Check if json formatted log data can be processed with the json_format flag."""
        json_data = b'{\n\t"a": 1,\n\t"b": {"x": 2}}'
        data = b'some log line.'
        any_dme = AnyByteDataModelElement('s')
        byte_stream_line_atomizer = ByteStreamLineAtomizer(any_dme, [], [self.stream_printer_event_handler], 300, [], json_format=True)
        self.assertEqual(byte_stream_line_atomizer.consume_data(json_data + data, False), len(json_data))

        # this is no valid json and should process only data until the last \n
        self.assertEqual(byte_stream_line_atomizer.consume_data(data + json_data + data, False), len(data) + json_data.rfind(b'\n') + 1)

        json_data = b'{"a": 1, "b": {"c": 2}, "d": 3}\n{"a": 1, "b": {"c": 2}, "d": 3}'
        self.assertEqual(byte_stream_line_atomizer.consume_data(json_data, False), len(json_data))
        self.assertEqual(byte_stream_line_atomizer.consume_data(json_data + data, False), len(json_data))

        json_data = b'{\n\t"a": 1,\n\t"b": {\n\t\t"c": 2},\n\t"d": 3}{\n"a": 1,\n\t"b": {\n\t\t"c": 2},\n\t"d": 3}'
        self.assertEqual(byte_stream_line_atomizer.consume_data(json_data + data, False), len(json_data))
        self.assertEqual(byte_stream_line_atomizer.consume_data(data + json_data, False), len(data) + json_data.rfind(b'\n') + 1)

        # even when the first json data gets invalidated, the second one starts after an empty line and is therefore valid until the end.
        json_data = b'{\n\t"a": 1,\n\t"b": {\n\t\t"c": 2},\n\t"d": 3}\n\n{\n"a": 1,\n\t"b": {\n\t\t"c": 2},\n\t"d": 3}'
        self.assertEqual(byte_stream_line_atomizer.consume_data(json_data + data, False), len(json_data))
        self.assertEqual(byte_stream_line_atomizer.consume_data(data + json_data + data, False), len(data) + len(json_data))

        # this is an incomplete json, but it still can be valid.
        json_data = b'{"a": 1, "b": {"c": 2}, "d": 3}\n{"a": 1, "b": {"c": 2}, "d'
        self.assertEqual(byte_stream_line_atomizer.consume_data(json_data, False), json_data.rfind(b'\n') + 1)

        # this is an incomplete json and the end can not be valid.
        json_data = b'{"a": 1, "b": {"c": 2}, "d": 3}\n{"a": 1, "b": {"c": 2}, d'
        self.assertEqual(byte_stream_line_atomizer.consume_data(json_data, False), json_data.rfind(b'\n') + 1)

    def test7json_max_line_length(self):
        """Check if json data is not parsed over the max_line_length."""
        json_data = b'{\n\t"a": 1,\n\t"b": {\n\t\t"c": 2},\n\t"d": 3}\n{\n"a": 1,\n\t"b": {"c": 2},"d": 3}\n'
        single_line_json_data = b'{"a": 1,"b": {"c": 2},"d": 3}{"a": 1,"b": {"c": 2},"d": 3'
        any_dme = AnyByteDataModelElement('s')
        byte_stream_line_atomizer = ByteStreamLineAtomizer(any_dme, [], [self.stream_printer_event_handler], 25, [], json_format=True)
        self.assertEqual(byte_stream_line_atomizer.consume_data(json_data, False), json_data.rfind(b'\n') + 1)
        self.assertEqual(self.output_stream.getvalue(), '')
        self.assertEqual(byte_stream_line_atomizer.consume_data(json_data, True), len(json_data))
        self.assertEqual(self.output_stream.getvalue(), '')
        byte_stream_line_atomizer = ByteStreamLineAtomizer(any_dme, [], [self.stream_printer_event_handler], 100, [], json_format=True)
        self.assertEqual(byte_stream_line_atomizer.consume_data(json_data, False), len(json_data))
        self.assertEqual(self.output_stream.getvalue(), '')
        self.assertEqual(byte_stream_line_atomizer.consume_data(json_data, True), len(json_data))
        self.assertEqual(self.output_stream.getvalue(), '')

        byte_stream_line_atomizer = ByteStreamLineAtomizer(any_dme, [], [self.stream_printer_event_handler], 25, [], json_format=True)
        self.assertEqual(byte_stream_line_atomizer.consume_data(single_line_json_data, True), len(single_line_json_data))
        self.assertEqual(self.output_stream.getvalue(), 'Overlong line terminated by end of stream (1 lines)\n  {"a": 1,"b": {"c": 2},"d":'
                                                        ' 3}{"a": 1,"b": {"c": 2},"d": 3\n\n')
        self.reset_output_stream()
        self.assertEqual(byte_stream_line_atomizer.consume_data(single_line_json_data, False), len(single_line_json_data))
        self.assertEqual(self.output_stream.getvalue(), '')
        byte_stream_line_atomizer = ByteStreamLineAtomizer(any_dme, [], [self.stream_printer_event_handler], 100, [], json_format=True)
        self.assertEqual(byte_stream_line_atomizer.consume_data(single_line_json_data, True), len(single_line_json_data))
        self.assertEqual(self.output_stream.getvalue(), 'Incomplete last line (1 lines)\n  {"a": 1,"b": {"c": 2},"d": 3\n\n')
        self.reset_output_stream()
        self.assertEqual(byte_stream_line_atomizer.consume_data(single_line_json_data, False), len(
            single_line_json_data.rsplit(b'}', 2)[0]) + 1)
        self.assertEqual(self.output_stream.getvalue(), '')


if __name__ == "__main__":
    unittest.main()



"""This module defines the parser for ulogd messages."""

from aminer.parsing.DecimalIntegerValueModelElement import DecimalIntegerValueModelElement
from aminer.parsing.DelimitedDataModelElement import DelimitedDataModelElement
from aminer.parsing.FirstMatchModelElement import FirstMatchModelElement
from aminer.parsing.FixedDataModelElement import FixedDataModelElement
from aminer.parsing.FixedWordlistDataModelElement import FixedWordlistDataModelElement
from aminer.parsing.IpAddressDataModelElement import IpAddressDataModelElement
from aminer.parsing.OptionalMatchModelElement import OptionalMatchModelElement
from aminer.parsing.SequenceModelElement import SequenceModelElement


def get_model():
    """Return a model for su session information messages after any standard logging preamble, e.g. from syslog."""
    type_children = [
        SequenceModelElement("build-stack", [
            FixedDataModelElement("s0", b"building new pluginstance stack: \""),
            DelimitedDataModelElement("stack", b"\""),
            FixedDataModelElement("s1", b"\"")
        ]),
        SequenceModelElement("nfct-event", [
            FixedDataModelElement("s0", b"[DESTROY] ORIG: SRC="),
            IpAddressDataModelElement("osrcip"),
            FixedDataModelElement("s1", b" DST="),
            IpAddressDataModelElement("odstip"),
            FixedDataModelElement("s2", b" PROTO="),
            FixedWordlistDataModelElement("proto", [b"TCP", b"UDP"]),
            FixedDataModelElement("s3", b" SPT="),
            DecimalIntegerValueModelElement("ospt"),
            FixedDataModelElement("s4", b" DPT="),
            DecimalIntegerValueModelElement("odpt"),
            FixedDataModelElement("s5", b" PKTS="),
            DecimalIntegerValueModelElement("opkts"),
            FixedDataModelElement("s6", b" BYTES="),
            DecimalIntegerValueModelElement("obytes"),
            FixedDataModelElement("s7", b" , REPLY: SRC="),
            IpAddressDataModelElement("rsrcip"),
            FixedDataModelElement("s8", b" DST="),
            IpAddressDataModelElement("rdstip"),
            FixedDataModelElement("s9", b" PROTO="),
            FixedWordlistDataModelElement("rproto", [b"TCP", b"UDP"]),
            FixedDataModelElement("s10", b" SPT="),
            DecimalIntegerValueModelElement("rspt"),
            FixedDataModelElement("s11", b" DPT="),
            DecimalIntegerValueModelElement("rdpt"),
            FixedDataModelElement("s12", b" PKTS="),
            DecimalIntegerValueModelElement("rpkts"),
            FixedDataModelElement("s13", b" BYTES="),
            DecimalIntegerValueModelElement("rbytes"),
            # No additional whitespace from Ubuntu Trusty 14.04 on.
            OptionalMatchModelElement("tail", FixedDataModelElement("s0", b" "))
        ]),
        FixedDataModelElement("nfct-plugin", b"NFCT plugin working in event mode"),
        FixedDataModelElement("reopen", b"reopening capture file"),
        FixedDataModelElement("signal", b"signal received, calling pluginstances"),
        FixedDataModelElement("uidchange", b"Changing UID / GID"),
        SequenceModelElement("seq", [
            FixedDataModelElement("s0", b"id=\""),
            DecimalIntegerValueModelElement("id"),
            FixedDataModelElement("s1", b"\" severity=\""),
            DelimitedDataModelElement("severity", b"\""),
            FixedDataModelElement("s2", b"\" sys=\""),
            DelimitedDataModelElement("sys", b"\""),
            FixedDataModelElement("s3", b"\" sub=\""),
            DelimitedDataModelElement("sub", b"\""),
            FixedDataModelElement("s4", b"\" name=\""),
            DelimitedDataModelElement("name", b"\""),
            FixedDataModelElement("s5", b"\" action=\""),
            DelimitedDataModelElement("action", b"\""),
            FixedDataModelElement("s6", b"\" fwrule=\""),
            DelimitedDataModelElement("fwrule", b"\""),
            FixedDataModelElement("s7", b"\" initf=\""),
            DelimitedDataModelElement("initf", b"\""),
            FixedDataModelElement("s8", b"\" srcmac=\""),
            DelimitedDataModelElement("srcmac", b"\""),
            FixedDataModelElement("s9", b"\" dstmac=\""),
            DelimitedDataModelElement("dstmac", b"\""),
            FixedDataModelElement("s10", b"\" srcip=\""),
            DelimitedDataModelElement("srcip", b"\""),
            FixedDataModelElement("s11", b"\" dstip=\""),
            DelimitedDataModelElement("dstip", b"\""),
            FixedDataModelElement("s12", b"\" proto=\""),
            DelimitedDataModelElement("proto", b"\""),
            FixedDataModelElement("s13", b"\" length=\""),
            DelimitedDataModelElement("length", b"\""),
            FixedDataModelElement("s14", b"\" tos=\""),
            DelimitedDataModelElement("tos", b"\""),
            FixedDataModelElement("s15", b"\" prec=\""),
            DelimitedDataModelElement("prec", b"\""),
            FixedDataModelElement("s16", b"\" ttl=\""),
            DelimitedDataModelElement("ttl", b"\""),
            FixedDataModelElement("s17", b"\" srcport=\""),
            DelimitedDataModelElement("srcport", b"\""),
            FixedDataModelElement("s18", b"\" dstport=\""),
            DelimitedDataModelElement("dstport", b"\""),
            FixedDataModelElement("s19", b"\" tcpflags=\""),
            DelimitedDataModelElement("tcpflags", b"\""),
            FixedDataModelElement("s20", b"\"")
        ])
    ]

    # Netflow entry
    model = SequenceModelElement("ulogd", [
        FixedDataModelElement("sname", b"ulogd["),
        DecimalIntegerValueModelElement("pid"),
        FixedDataModelElement("s0", b"]: "),
        FirstMatchModelElement("msg", type_children)
    ])
    return model



import unittest
from aminer.parsing.MatchContext import MatchContext
from aminer.parsing.FixedDataModelElement import FixedDataModelElement
from aminer.analysis.NewMatchPathDetector import NewMatchPathDetector
from aminer.input.LogAtom import LogAtom
from aminer.parsing.ParserMatch import ParserMatch
from aminer.analysis.TimestampCorrectionFilters import SimpleMonotonicTimestampAdjust
from time import time
from unit.TestBase import TestBase
from datetime import datetime


class TimestampCorrectionFiltersTest(TestBase):
    """Unittests for the TimestampCorrectionFilters."""

    __expected_string = '%s New path(es) detected\n%s: "%s" (%d lines)\n  %s\n pid=\n\n'
    match_path = "['match/s1']"

    def test1simple_monotonic_timestamp_adjust_test(self):
        """This test case checks if the timestamp is adjusted and logAtoms are forwarded correctly."""
        description = "Test1TimestampCorrectionFilter"
        match_context_fixed_dme = MatchContext(b' pid=')
        fixed_dme = FixedDataModelElement('s1', b' pid=')
        match_element_fixed_dme = fixed_dme.get_match_element("match", match_context_fixed_dme)
        t = time()

        new_match_path_detector = NewMatchPathDetector(self.aminer_config, [self.stream_printer_event_handler], 'Default', False,
                                                       output_logline=False)
        self.analysis_context.register_component(new_match_path_detector, description)
        simple_monotonic_timstamp_adjust = SimpleMonotonicTimestampAdjust([new_match_path_detector], False)
        log_atom_fixed_dme = LogAtom(fixed_dme.fixed_data, ParserMatch(match_element_fixed_dme), t, new_match_path_detector)
        self.assertEqual(simple_monotonic_timstamp_adjust.receive_atom(log_atom_fixed_dme), True)
        self.assertEqual(self.output_stream.getvalue(), self.__expected_string % (
            datetime.fromtimestamp(t).strftime("%Y-%m-%d %H:%M:%S"), new_match_path_detector.__class__.__name__, description, 1,
            self.match_path))


if __name__ == "__main__":
    unittest.main()



"""
This file defines the EnhancedNewMatchPathValueComboDetector.
detector to extract values from LogAtoms and check, if the value
combination was already seen before.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""

import time
import os
import logging

from aminer.analysis.NewMatchPathValueComboDetector import NewMatchPathValueComboDetector
from aminer.util import PersistenceUtil
from aminer.AminerConfig import DEBUG_LOG_NAME, KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD, STAT_LOG_NAME,\
    CONFIG_KEY_LOG_LINE_PREFIX, DEFAULT_LOG_LINE_PREFIX
from aminer import AminerConfig


class EnhancedNewMatchPathValueComboDetector(NewMatchPathValueComboDetector):
    """
    This class creates events when a new value combination for a given list of match data paths were found.
    It is similar to the NewMatchPathValueComboDetector basic detector but also provides support for storing meta information about each
    detected value combination, e.g.
    * the first time a tuple was detected using the LogAtom default timestamp.
    * the last time a tuple was seen
    * the number of times the tuple was seen
    * user data for annotation.
    Due to the additional features, this detector is slower than the basic detector.
    """

    def __init__(self, aminer_config, target_path_list, anomaly_event_handlers, persistence_id='Default', allow_missing_values_flag=False,
                 learn_mode=False, tuple_transformation_function=None, output_logline=True, stop_learning_time=None,
                 stop_learning_no_anomaly_time=None):
        """
        Initialize the detector. This will also trigger reading or creation of persistence storage location.
        @param target_path_list the list of values to extract from each match to create the value combination to be checked.
        @param allow_missing_values_flag when set to True, the detector will also use matches, where one of the paths from target_path_list
               does not refer to an existing parsed data object.
        @param learn_mode when set to True, this detector will report a new value only the first time before including it in the known
               values set automatically.
        @param tuple_transformation_function when not None, this function will be invoked on each extracted value combination list to
               transform it. It may modify the list directly or create a new one to return it.
        @param stop_learning_time switch the learn_mode to False after the time.
        @param stop_learning_no_anomaly_time switch the learn_mode to False after no anomaly was detected for that time.
        """
        # avoid "defined outside init" issue
        self.learn_mode, self.stop_learning_timestamp, self.next_persist_time, self.log_success, self.log_total = [None]*5
        self.known_values_dict = {}
        self.tuple_transformation_function = tuple_transformation_function
        super().__init__(
            aminer_config=aminer_config, target_path_list=target_path_list, anomaly_event_handlers=anomaly_event_handlers,
            persistence_id=persistence_id, allow_missing_values_flag=allow_missing_values_flag, learn_mode=learn_mode,
            output_logline=output_logline, stop_learning_time=stop_learning_time,
            stop_learning_no_anomaly_time=stop_learning_no_anomaly_time)
        self.date_string = "%Y-%m-%d %H:%M:%S"
        self.log_learned_path_value_combos = 0
        self.log_new_learned_values = []

    def load_persistence_data(self):
        """Load the persistence data from storage."""
        persistence_data = PersistenceUtil.load_json(self.persistence_file_name)
        if persistence_data is not None:
            # Dictionary and tuples were stored as list of lists. Transform
            # the first lists to tuples to allow hash operation needed by set.
            for value_tuple, extra_data in persistence_data:
                self.known_values_dict[tuple(value_tuple)] = extra_data
            logging.getLogger(DEBUG_LOG_NAME).debug(f'{self.__class__.__name__} loaded persistence data.')

    def receive_atom(self, log_atom):
        """
        Receive on parsed atom and the information about the parser match.
        @return True if a value combination was extracted and checked against the list of known combinations, no matter if the checked
        values were new or not.
        """
        self.log_total += 1
        match_dict = log_atom.parser_match.get_match_dictionary()
        if self.learn_mode is True and self.stop_learning_timestamp is not None and \
                self.stop_learning_timestamp < log_atom.atom_time:
            logging.getLogger(DEBUG_LOG_NAME).info(f"Stopping learning in the {self.__class__.__name__}.")
            self.learn_mode = False
        timestamp = log_atom.get_timestamp()
        if timestamp is None:
            timestamp = time.time()
        timestamp = round(timestamp, 3)
        match_value_list = []
        for target_path in self.target_path_list:
            match = match_dict.get(target_path)
            if match is None:
                if not self.allow_missing_values_flag:
                    return False
                match_value_list.append(None)
            else:
                matches = []
                if isinstance(match, list):
                    matches = match
                else:
                    matches.append(match)
                for match_element in matches:
                    match_value_list.append(match_element.match_object)

        if self.tuple_transformation_function is not None:
            match_value_list = self.tuple_transformation_function(match_value_list)
        match_value_tuple = tuple(match_value_list)

        if self.known_values_dict.get(match_value_tuple) is None:
            self.known_values_dict[match_value_tuple] = [timestamp, timestamp, 1]
            self.log_new_learned_values.append(match_value_tuple)
        else:
            extra_data = self.known_values_dict.get(match_value_tuple)
            extra_data[1] = timestamp
            extra_data[2] += 1

        affected_log_atom_values = []
        metadata = {}
        for match_value in list(match_value_tuple):
            if isinstance(match_value, bytes):
                match_value = match_value.decode(AminerConfig.ENCODING)
            affected_log_atom_values.append(str(match_value))
        values = self.known_values_dict.get(match_value_tuple)
        metadata['TimeFirstOccurrence'] = values[0]
        metadata['TimeLastOccurrence'] = values[1]
        metadata['NumberOfOccurrences'] = values[2]

        analysis_component = {'AffectedLogAtomPaths': self.target_path_list, 'AffectedLogAtomValues': affected_log_atom_values,
                              'Metadata': metadata}
        event_data = {'AnalysisComponent': analysis_component}
        if (self.learn_mode and self.known_values_dict.get(match_value_tuple)[2] == 1) or not self.learn_mode:
            self.log_learned_path_value_combos += 1
            try:
                data = log_atom.raw_data.decode(AminerConfig.ENCODING)
            except UnicodeError:
                data = repr(log_atom.raw_data)
            original_log_line_prefix = self.aminer_config.config_properties.get(CONFIG_KEY_LOG_LINE_PREFIX, DEFAULT_LOG_LINE_PREFIX)
            for listener in self.anomaly_event_handlers:
                sorted_log_lines = [str(self.known_values_dict) + os.linesep + original_log_line_prefix + data]
                listener.receive_event(f'Analysis.{self.__class__.__name__}', 'New value combination(s) detected', sorted_log_lines,
                                       event_data, log_atom, self)
            if self.learn_mode and self.stop_learning_timestamp is not None and self.stop_learning_no_anomaly_time is not None:
                self.stop_learning_timestamp = time.time() + self.stop_learning_no_anomaly_time
        self.log_success += 1
        return True

    def do_timer(self, trigger_time):
        """Check if current ruleset should be persisted."""
        if self.next_persist_time is None:
            return self.aminer_config.config_properties.get(KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD)

        delta = self.next_persist_time - trigger_time
        if delta <= 0:
            self.do_persist()
            delta = self.aminer_config.config_properties.get(KEY_PERSISTENCE_PERIOD, DEFAULT_PERSISTENCE_PERIOD)
            self.next_persist_time = time.time() + delta
        return delta

    def do_persist(self):
        """Immediately write persistence data to storage."""
        persistence_data = []
        for dict_record in self.known_values_dict.items():
            persistence_data.append(dict_record)
        PersistenceUtil.store_json(self.persistence_file_name, persistence_data)
        logging.getLogger(DEBUG_LOG_NAME).debug(f'{self.__class__.__name__} persisted data.')

    def allowlist_event(self, event_type, event_data, allowlisting_data):
        """
        Allowlist an event generated by this source using the information emitted when generating the event.
        @return a message with information about allowlisting
        @throws Exception when allowlisting of this special event using given allowlisting_data was not possible.
        """
        if event_type != f'Analysis.{self.__class__.__name__}':
            msg = 'Event not from this source'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        if allowlisting_data is not None:
            msg = 'Allowlisting data not understood by this detector'
            logging.getLogger(DEBUG_LOG_NAME).error(msg)
            raise Exception(msg)
        current_timestamp = event_data[0]
        self.known_values_dict[event_data[1]] = [current_timestamp, current_timestamp, 1]
        return f"Allowlisted path(es) {', '.join(self.target_path_list)} with {event_data}."

    def log_statistics(self, component_name):
        """
        Log statistics of an AtomHandler. Override this method for more sophisticated statistics output of the AtomHandler.
        @param component_name the name of the component which is printed in the log line.
        """
        if AminerConfig.STAT_LEVEL == 1:
            logging.getLogger(STAT_LOG_NAME).info(
                f"'{component_name}' processed {self.log_success} out of {self.log_total} log atoms successfully and learned "
                f"{self.log_learned_path_value_combos} new value combinations in the last 60 minutes.")
        elif AminerConfig.STAT_LEVEL == 2:
            logging.getLogger(STAT_LOG_NAME).info(
                f"'{component_name}' processed {self.log_success} out of {self.log_total} log atoms successfully and learned "
                f"{self.log_learned_path_value_combos} new value combinations in the last 60 minutes. Following new value combinations "
                f"were learned: {self.log_new_learned_values}")
        self.log_success = 0
        self.log_total = 0
        self.log_learned_path_value_combos = 0
        self.log_new_learned_values = []



"""
This module defines an event handler that prints data to a local syslog instance.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""

import io
import os
import syslog

from aminer.events.EventInterfaces import EventHandlerInterface
from aminer.events.StreamPrinterEventHandler import StreamPrinterEventHandler


class SyslogWriterEventHandler(EventHandlerInterface):
    """
    This class implements an event record listener to forward events to the local syslog instance.
    CAVEAT: USE THIS AT YOUR OWN RISK: by creating aminer/syslog log data processing loops, you will flood your syslog and probably
    fill up your disks.
    """

    def __init__(self, analysis_context, instance_name='aminer'):
        """
        Initialize the event handler.
        @param analysis_context the analysis context used to get the component.
        @param instance_name the process name shown in the syslog.
        """
        self.analysis_context = analysis_context
        self.instanceName = instance_name
        syslog.openlog(f'{self.instanceName}[{os.getpid()}]', syslog.LOG_INFO, syslog.LOG_DAEMON)
        syslog.syslog(syslog.LOG_INFO, 'Syslog logger initialized')
        self.buffer_stream = io.StringIO()
        self.event_writer = StreamPrinterEventHandler(analysis_context, self.buffer_stream)
        self.event_id = 0

    def receive_event(self, event_type, event_message, sorted_loglines, event_data, log_atom, event_source):
        """
        Receive information about a detected even and forward it to syslog.
        @param event_type is a string with the event type class this event belongs to. This information can be used to interpret
               type-specific event_data objects. Together with the eventMessage and sorted_loglines, this can be used to create generic log
               messages.
        @param event_message the first output line of the event.
        @param sorted_loglines sorted list of log lines that were considered when generating the event, as far as available to the time
               of the event. The list has to contain at least one line.
        @param event_data type-specific event data object, should not be used unless listener really knows about the event_type.
        @param log_atom the log atom which produced the event.
        @param event_source reference to detector generating the event.
        """
        if hasattr(event_source, 'output_event_handlers') and event_source.output_event_handlers is not None and self not in \
                event_source.output_event_handlers:
            return
        component_name = self.analysis_context.get_name_by_component(event_source)
        if component_name in self.analysis_context.suppress_detector_list:
            return
        self.buffer_stream.seek(0)
        self.buffer_stream.truncate(0)
        self.event_writer.receive_event(event_type, event_message, sorted_loglines, event_data, log_atom, event_source)
        event_data = self.buffer_stream.getvalue()
        current_event_id = self.event_id
        self.event_id += 1
        serial = 0
        for data_line in event_data.strip().split('\n'):
            # Python syslog is very ugly if lines are too long, so break them down.
            while data_line:
                message = None
                if serial == 0:
                    message = f'[{current_event_id}] {data_line[:800]}'
                else:
                    message = f'[{current_event_id}-{serial}] {data_line[:800]}'
                data_line = data_line[800:]
                syslog.syslog(syslog.LOG_INFO, message)
                serial += 1
